from typing import Dict, Type, List

from planqk.braket.aws_device import PlanqkAwsDevice
from planqk.client.backend_dtos import BackendDto
from planqk.exceptions import BackendNotFoundError, PlanqkClientError
from planqk.qiskit.provider import _PlanqkProvider


class PlanqkBraketProvider(_PlanqkProvider):
    _device_mapping: Dict[str, Type[PlanqkAwsDevice]] = {}

    @classmethod
    def register_device(cls, device_id: str):
        """For internal use only. Binds a device class to a PLANQK backend id."""

        def decorator(device_cls: Type[PlanqkAwsDevice]):
            cls._device_mapping[device_id] = device_cls
            return device_cls

        return decorator

    @classmethod
    def get_device_class(cls, backend_id: str) -> Type[PlanqkAwsDevice]:
        device_class = cls._device_mapping.get(backend_id)
        if device_class is None:
            raise BackendNotFoundError(f"Backend '{backend_id}' is not supported by PLANQK Braket.")
        return device_class

    def get_device(self, backend_id: str) -> PlanqkAwsDevice:
        """
        Retrieves an AWS Braket Device based on the provided PLANQK backend id.

        Args:
            backend_id (str): The PLANQK backend (device) id.

        Returns:
            PlanqkAwsDevice: The AWS Braket device corresponding to the backend id.

        Raises:
            BackendNotFoundError: If the backend with the given id cannot be found or is not supported by the Braket SDK.

        Note:
            An overview of the supported backends and their IDs can be found at: https://platform.planqk.de/quantum-backends
        """
        backend_info = self._get_backend_info(backend_id)
        return self._get_planqk_braket_device(backend_info)

    def devices(self) -> List[str]:
        """
        Retrieves a list of all AWS Braket devices (backends) provided through PLANQK that can be accessed using this SDK.

        Returns:
            List[str]: A list of supported AWS Braket device IDs.
        """
        return list(self._device_mapping.keys())

    def _get_backend_info(self, backend_id):
        try:
            return self._client.get_backend(backend_id=backend_id)
        except PlanqkClientError as e:
            if e.response.status_code == 404:
                raise BackendNotFoundError(f"PLANQK device with id {backend_id} not found.")
            raise e

    def _get_planqk_braket_device(self, backend_info: BackendDto) -> PlanqkAwsDevice:
        device_class = self.get_device_class(backend_info.id)
        return device_class(planqk_client=self._client, backend_info=backend_info)



from typing import Optional

from planqk.backend import PlanqkBackend
from planqk.client.client import _PlanqkClient
from planqk.client.job_dtos import JobDto
from planqk.client.model_enums import Provider
from planqk.qiskit import PlanqkQiskitJob
from planqk.qiskit.providers.aws.aws_qiskit_job import PlanqkAwsQiskitJob
from planqk.qiskit.providers.azure.azure_qiskit_job import PlanqkAzureQiskitJob
from planqk.qiskit.providers.qryd.qryd_qiskit_job import PlanqkQrydQiskitJob
from planqk.qiskit.providers.qudora.qudora_sim_job import PlanqkQudoraQiskitJob


class PlanqkQiskitJobFactory:
    @staticmethod
    def create_job(backend: Optional[PlanqkBackend], job_id: Optional[str] = None, job_details: Optional[JobDto] = None,
                   planqk_client: Optional[_PlanqkClient] = None) -> PlanqkQiskitJob:

        provider = PlanqkQiskitJobFactory._get_provider(backend, job_details)

        if not provider:
            raise ValueError("Provider information is missing. Either 'backend' or 'job_details' with the 'provider' attribute must be specified.")

        if provider == Provider.AWS:
            return PlanqkAwsQiskitJob(backend, job_id, job_details, planqk_client)
        elif provider == Provider.AZURE:
            return PlanqkAzureQiskitJob(backend, job_id, job_details, planqk_client)
        elif provider == Provider.QRYD:
            return PlanqkQrydQiskitJob(backend, job_id, job_details, planqk_client)
        elif provider == Provider.QUDORA:
            return PlanqkQudoraQiskitJob(backend, job_id, job_details, planqk_client)
        else:
            return PlanqkQiskitJob(backend, job_id, job_details, planqk_client)

    @staticmethod
    def _get_provider(backend, job_details):
        if backend:
            provider = backend.backend_info.provider
        else:
            provider = job_details.provider if job_details else None
        return provider


from __future__ import annotations

import json
from typing import Any, Union, Optional

from braket.ahs.analog_hamiltonian_simulation import AnalogHamiltonianSimulation
from braket.annealing.problem import Problem
from braket.aws import AwsQuantumTask
from braket.aws.aws_session import AwsSession
from braket.aws.queue_information import QuantumTaskQueueInfo
from braket.circuits.circuit import Circuit, Gate, QubitSet
from braket.ir.blackbird import Program as BlackbirdProgram
from braket.ir.openqasm import Program as OpenQASMProgram
from braket.pulse.pulse_sequence import PulseSequence
from braket.schema_common import BraketSchemaBase
from braket.tasks import AnalogHamiltonianSimulationQuantumTaskResult, GateModelQuantumTaskResult

from planqk.client.client import _PlanqkClient
from planqk.client.job_dtos import JobDto
from planqk.client.model_enums import PlanqkJobStatus
from planqk.decorators import not_implemented
from planqk.exceptions import PlanqkError
from planqk.job import PlanqkBaseJob


class PlanqkAwsQuantumTask(PlanqkBaseJob, AwsQuantumTask):

    def __init__(self,
                 task_id: Optional[str] = None,
                 access_token: Optional[str] = None,
                 organization_id: Optional[str] = None,
                 _job_details: Optional[JobDto] = None,
                 _backend: Optional = None,
                 _client: _PlanqkClient = None):
        """
        Initialize the PlanqkAwsQuantumTask.

        Args:
            task_id (str, optional):
                The unique identifier of the quantum task. This ID is used to reference an existing task on PLANQK,
                allowing the task's status, results, and other details to be retrieved or managed.
                Defaults to None.

            access_token (str, optional):
                Access token used for authentication with PLANQK. If no token is provided, the token is retrieved
                from the environment variable `PLANQK_ACCESS_TOKEN`, which can be set manually or by using the
                PLANQK CLI. This token is used to authorize access to PLANQK services. Defaults to None.

            organization_id (str, optional):
                The ID of a PLANQK organization you are a member of. Provide this ID if you want to access
                quantum backends with an organization account and its associated pricing plan. All backend
                executions (jobs, tasks, etc.) you create are visible to the members of the organization.
                If the ID is omitted, all backend executions are performed under your personal account.
                Defaults to None.

            _job_details (JobDto, optional):
                Internal use only. Contains detailed information about the details associated with this task,
                including metadata such as name, status, and configuration details relevant to the task.
                Defaults to None.

            _backend (optional):
                Internal use only. Specifies the backend on which the quantum task is executed.
                Defaults to None.

            _client (_PlanqkClient, optional):
                Internal use only. A client instance used for making requests to the PLANQK API. This parameter is
                mainly intended for testing purposes.
                Defaults to None.

         Raises:
            BackendNotFoundError: If the device where the task was executed on cannot be found or is not supported by the Braket SDK.
        """
        client = _client or _PlanqkClient(access_token=access_token, organization_id=organization_id)
        PlanqkBaseJob.__init__(self, backend=_backend, job_id=task_id, job_details=_job_details, planqk_client=client)

        is_existing_task = task_id is not None
        if is_existing_task:
            self._verify_task_device_is_supported()

    def _verify_task_device_is_supported(self):
        from planqk import PlanqkBraketProvider
        backend_id = self._job_details.backend_id
        PlanqkBraketProvider.get_device_class(backend_id)

    @staticmethod
    @not_implemented
    def create(
        aws_session: AwsSession,
        device_arn: str,
        task_specification: Union[
            Circuit,
            Problem,
            OpenQASMProgram,
            BlackbirdProgram,
            PulseSequence,
            AnalogHamiltonianSimulation,
        ],
        s3_destination_folder: AwsSession.S3DestinationFolder,
        shots: int,
        device_parameters: dict[str, Any] | None = None,
        disable_qubit_rewiring: bool = False,
        tags: dict[str, str] | None = None,
        inputs: dict[str, float] | None = None,
        gate_definitions: dict[tuple[Gate, QubitSet], PulseSequence] | None = None,
        quiet: bool = False,
        reservation_arn: str | None = None,
        *args,
        **kwargs,
    ) -> AwsQuantumTask:
        # Task must be either created with PlanqkAwsDevice.run() or retrieved by specifying the task ID in the constructor of the PlanqkAwsQuantumTask class.
        pass

    @not_implemented
    def metadata(self, use_cached_value: bool = False) -> dict[str, Any]:
        """Get quantum task metadata defined in Amazon Braket.

        Args:
            use_cached_value (bool): If `True`, uses the value most recently retrieved
                from the Amazon Braket `GetQuantumTask` operation, if it exists; if not,
                `GetQuantumTask` will be called to retrieve the metadata. If `False`, always calls
                `GetQuantumTask`, which also updates the cached value. Default: `False`.

        Returns:
            dict[str, Any]: The response from the Amazon Braket `GetQuantumTask` operation.
            If `use_cached_value` is `True`, Amazon Braket is not called and the most recently
            retrieved value is used, unless `GetQuantumTask` was never called, in which case
            it will still be called to populate the metadata for the first time.
        """
        pass

    def state(self, use_cached_value: bool = False) -> str:
        """The state of the quantum task.

        Args:
            use_cached_value (bool): If `True`, uses the value most recently retrieved from PLANQK.

        Returns:
            str: the job execution state.
        """
        state = PlanqkBaseJob._update_state(self, use_cached_value)
        return 'QUEUED' if state == PlanqkJobStatus.PENDING else state.value

    @not_implemented
    def queue_position(self) -> QuantumTaskQueueInfo:
        """The queue position details for the quantum task."""
        pass

    @property
    def id(self) -> str:
        """Get the quantum task ID.

        Returns:
            str: The quantum task ID.
        """
        return super().id

    def job_id(self) -> str:
        """Get the quantum task ID.

        Returns:
            str: The quantum task ID.
        """
        return self.id

    def cancel(self) -> None:
        """Cancel the quantum task."""
        super().cancel()

    @not_implemented
    def async_result(self):
        """Get the quantum task result asynchronously.

        Returns:
            asyncio.Task: Get the quantum task result asynchronously.
        """
        pass

    def result(
        self,
    ) -> Union[AnalogHamiltonianSimulationQuantumTaskResult, GateModelQuantumTaskResult]:
        """Retrieve the quantum task result by polling PLANQK until the task is completed.

        Returns:
            Union[AnalogHamiltonianSimulationQuantumTaskResult, GateModelQuantumTaskResult]:
                An instance of `AnalogHamiltonianSimulationQuantumTaskResult` if the task was executed on an Analog Hamiltonian Simulator
                or an instance of `GateModelQuantumTaskResult` if the task was performed on a Gate-based device.
        Raises:
            PlanqkError: If the AWS task result type is unexpected or unrecognized.
        """
        try:
            result = super()._result()
            result_type = result.get('braketSchemaHeader', {}).get('name')
            if result_type == "braket.task_result.analog_hamiltonian_simulation_task_result":
                return AnalogHamiltonianSimulationQuantumTaskResult.from_string(json.dumps(result))

            elif result_type == "braket.task_result.gate_model_task_result":
                gate_model_result = BraketSchemaBase.parse_raw_schema(json.dumps(result))
                GateModelQuantumTaskResult.cast_result_types(gate_model_result)
                return GateModelQuantumTaskResult.from_object(gate_model_result)
            else:
                raise ValueError(f"Unexpected AWS task result type '{result_type}'")
        except Exception as e:
            raise PlanqkError(f"Cannot process AWS task result {result}: {e}") from e



import json
from abc import ABC
from typing import List, Dict, Type

from planqk.client.backend_dtos import BackendDto
from planqk.client.client import _PlanqkClient
from planqk.client.job_dtos import JobSummary
from planqk.client.model_enums import Provider
from planqk.exceptions import PlanqkClientError, BackendNotFoundError, PlanqkError
from planqk.qiskit.backend import PlanqkQiskitBackend
from planqk.qiskit.job_factory import PlanqkQiskitJobFactory


class _PlanqkProvider(ABC):
    def __init__(self, access_token: str = None, organization_id: str = None, _client=None):
        """Initialize the PLANQK provider.
              Args:
                    access_token (str): access token used for authentication with PLANQK.
                        If no token is provided, the token is retrieved from the environment variable PLANQK_ACCESS_TOKEN that can be either set
                        manually or by using the PLANQK CLI.
                        Defaults to None.

                    organization_id (str, optional): the ID of a PLANQK organization you are member of.
                        Provide this ID if you want to access quantum
                        backends with an organization account and its associated pricing plan.
                        All backend executions (jobs, tasks etc.) you create are visible to the members of the organization.
                        If the ID is omitted, all backend executions are performed under your personal account.
                        Defaults to None.

                    _client (_PlanqkClient, optional): Client instance used for making requests to the PLANQK API.
                        This parameter is mainly intended for testing purposes.
                        Defaults to None.
        """
        self._client = _client or _PlanqkClient(access_token=access_token, organization_id=organization_id)


class PlanqkQuantumProvider(_PlanqkProvider):
    _backend_mapping: Dict[str, Type[PlanqkQiskitBackend]] = {}

    @classmethod
    def register_backend(cls, backend_id: str):
        """For internal use only. Binds a backend class to a PLANQK backend id."""

        def decorator(backend_cls: Type[PlanqkQiskitBackend]):
            cls._backend_mapping[backend_id] = backend_cls
            return backend_cls

        return decorator

    def backends(self, provider: Provider = None) -> List[str]:
        """Return the list of backend IDs supported by PLANQK.

       Args:
           provider: returns only the IDs of the backend of the given provider, if specified. Defaults to None.

       Returns:
           List[str]: a list of backend ids that match the filtering criteria.
        """
        backend_dtos = self._client.get_backends()

        supported_backend_ids = [
            backend_info.id for backend_info in backend_dtos
            if (provider is None or backend_info.provider == provider) and backend_info.provider != Provider.DWAVE
        ]
        return supported_backend_ids

    def get_backend(self, backend_id: str) -> PlanqkQiskitBackend:
        """Return a single backend matching the specified filtering.

        Args:
            backend_id: name of the backend.
            sdk: the SDK to use. Note, that backends must be supported by the SDK.
            **kwargs: dict used for filtering.

        Returns:
            Backend: a backend matching the filtering criteria.

        Raises:
            BackendNotFoundError: if no backend could be found or more than one backend matches the filtering criteria.
        """
        try:
            backend_dto = self._client.get_backend(backend_id=backend_id)
        except PlanqkClientError as e:
            if e.response.status_code == 404:
                text = e.response.text
                if text:
                    error_detail = json.loads(e.response.text)
                    raise BackendNotFoundError("No backend matches the criteria. Reason: " + error_detail['error'])
                else:
                    raise BackendNotFoundError("No backend matches the criteria.")
            raise e

        backend_state_dto = self._client.get_backend_state(backend_id=backend_id)
        if backend_state_dto:
            backend_dto.status = backend_state_dto.status

        return self._get_backend_object(backend_dto)

    def _get_backend_object(self, backend_info: BackendDto) -> PlanqkQiskitBackend:
        backend_class = self._get_backend_class(backend_info.id)

        if backend_class:
            return backend_class(planqk_client=self._client, backend_info=backend_info)
        else:
            raise BackendNotFoundError(f"Qiskit backend '{backend_info.id}' is not supported.")

    def _get_backend_class(self, backend_id: str) -> Type[PlanqkQiskitBackend]:
        return self._backend_mapping.get(backend_id)

    def retrieve_job(self, job_id: str):
        """
        Retrieve a job from the backend.

        Args:
            backend (PlanqkQiskitBackend): the backend that run the job.
            job_id (str): the job id.

        Returns:
            Job: the job from the backend with the given id.
        """
        job_details = self._client.get_job(job_id=job_id)
        backend_id = job_details.backend_id
        backend_class = self._get_backend_class(backend_id)
        if backend_class is None:
            raise PlanqkError(f"Job '{job_id}' was created on backend '{backend_id}'. This backend is not supported by Qiskit.")

        return PlanqkQiskitJobFactory.create_job(backend=None, job_id=job_id, job_details=job_details, planqk_client=self._client)

    def jobs(self) -> List[JobSummary]:
        """
        Returns an overview on all jobs of the user or an organization.

        For each job the following attributes are returned:
            - job_id (str): The unique identifier of the job.
            - provider_name (Provider): The job provider.
            - backend_id (str): The identifier of the backend used.
            - created_at (datetime): The timestamp when the job was created.

        Returns:
            List[JobSummary]: A list of basic job information.
        """
        print("Getting your jobs from PLANQK, this may take a few seconds...")
        job_dtos = self._client.get_jobs()
        return [
            JobSummary(
                id=job_dto.id,
                provider=job_dto.provider,
                backend_id=job_dto.backend_id,
                created_at=job_dto.created_at
            )
            for job_dto in job_dtos
        ]



from enum import Enum


class Provider(Enum):
    AZURE = "AZURE"
    AWS = "AWS"
    DWAVE = "DWAVE"
    IBM = "IBM"
    IBM_CLOUD = "IBM_CLOUD"
    TSYSTEMS = "TSYSTEMS"
    QRYD = "QRYD"
    QUDORA = "QUDORA"
    UNKNOWN = "UNKNOWN"

    @classmethod
    def from_str(cls, provider_str):
        try:
            return Provider(provider_str)
        except ValueError:
            return cls.UNKNOWN


class BackendType(Enum):
    QPU = "QPU"
    SIMULATOR = "SIMULATOR"
    ANNEALER = "ANNEALER"
    UNKNOWN = "UNKNOWN"

    @classmethod
    def from_str(cls, type_str):
        try:
            return BackendType(type_str)
        except ValueError:
            return cls.UNKNOWN


class HardwareProvider(Enum):
    IONQ = "IONQ"
    RIGETTI = "RIGETTI"
    OQC = "OQC"
    AWS = "AWS"
    AZURE = "AZURE"
    IBM = "IBM"
    QRYD = "QRYD"
    DWAVE = "DWAVE"
    QUERA = "QUERA"
    IQM = "IQM"
    QUDORA = "QUDORA"
    QUANTINUUM = "QUANTINUUM"
    UNKNOWN = "UNKNOWN"

    @classmethod
    def from_str(cls, hw_provider_str):
        try:
            return HardwareProvider(hw_provider_str)
        except ValueError:
            return cls.UNKNOWN


class BackendStatus(Enum):
    """
    STATUS Enum:

    UNKNOWN: The actual status is unknown.
    ONLINE: The actual is online, processing submitted jobs and accepting new ones.
    PAUSED: The actual is accepting jobs, but not currently processing them.
    OFFLINE: The actual is not accepting new jobs, e.g. due to maintenance.
    RETIRED: The actual is not available for use anymore.
    """
    UNKNOWN = "UNKNOWN"
    ONLINE = "ONLINE"
    PAUSED = "PAUSED"
    OFFLINE = "OFFLINE"
    RETIRED = "RETIRED"

    @classmethod
    def from_str(cls, status_str):
        try:
            return BackendStatus(status_str)
        except ValueError:
            return cls.UNKNOWN


class PlanqkSdkProvider(Enum):
    QISKIT = "QISKIT"
    BRAKET = "BRAKET"
    UNKNOWN = "UNKNOWN"

    @classmethod
    def from_str(cls, provider_str):
        try:
            return PlanqkSdkProvider(provider_str)
        except ValueError:
            return cls.UNKNOWN


class PlanqkJobStatus(str, Enum):
    UNKNOWN = "UNKNOWN"
    ABORTED = "ABORTED"
    PENDING = "PENDING"
    RUNNING = "RUNNING"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    CANCELLING = "CANCELLING"
    CANCELLED = "CANCELLED"

    @classmethod
    def from_str(cls, job_status_str):
        try:
            return PlanqkJobStatus(job_status_str.upper())
        except ValueError:
            return cls.UNKNOWN


JOB_FINAL_STATES = (PlanqkJobStatus.ABORTED, PlanqkJobStatus.COMPLETED, PlanqkJobStatus.CANCELLED, PlanqkJobStatus.FAILED)


class JobInputFormat(str, Enum):
    OPEN_QASM_V1 = "OPEN_QASM_V1"
    OPEN_QASM_V2 = "OPEN_QASM_V2"
    OPEN_QASM_V3 = "OPEN_QASM_V3"
    QIR_V1 = "QIR_V1"
    BRAKET_OPEN_QASM_V3 = "BRAKET_OPEN_QASM_V3"
    BRAKET_AHS_PROGRAM = "BRAKET_AHS_PROGRAM"
    IONQ_CIRCUIT_V1 = "IONQ_CIRCUIT_V1"
    QISKIT = "QISKIT"
    QOQO = "QOQO"
    UNKNOWN = "UNKNOWN"

    @classmethod
    def from_str(cls, job_input_format_str):
        try:
            return JobInputFormat(job_input_format_str)
        except ValueError:
            return cls.UNKNOWN



from datetime import time, date
from typing import Dict
from typing import List
from typing import Optional

from pydantic import BaseModel, field_validator

from planqk.client.dto_utils import init_with_defined_params
from planqk.client.model_enums import Provider, BackendType, HardwareProvider, BackendStatus, JobInputFormat


class DocumentationDto(BaseModel):
    description: Optional[str] = None
    url: Optional[str] = None
    # status_url: Optional[str] = None
    location: Optional[str] = None

    @classmethod
    def from_dict(cls, data: Dict):
        return init_with_defined_params(cls, data)


class QubitDto(BaseModel):
    id: str

    @classmethod
    def from_dict(cls, data: Dict):
        return cls(**data)


class GateDto(BaseModel):
    name: str
    native_gate: bool

    @classmethod
    def from_dict(cls, data: Dict):
        return init_with_defined_params(cls, data)


class ConnectivityDto(BaseModel):
    fully_connected: bool
    graph: Optional[Dict[str, List[str]]] = None

    @classmethod
    def from_dict(cls, data: Dict):
        return init_with_defined_params(cls, data)


class ShotsRangeDto(BaseModel):
    min: int
    max: int

    @classmethod
    def from_dict(cls, data: Dict):
        return init_with_defined_params(cls, data)


class ConfigurationDto(BaseModel):
    gates: List[GateDto]
    instructions: List[str]
    qubits: Optional[List[QubitDto]] = None
    qubit_count: int
    connectivity: Optional[ConnectivityDto] = None
    supported_input_formats: List[JobInputFormat]
    shots_range: ShotsRangeDto
    memory_result_supported: Optional[bool] = False
    options: Optional[Dict] = None

    def __post_init__(self):
        self.gates = [GateDto.from_dict(gate) for gate in self.gates]
        self.qubits = [QubitDto.from_dict(qubit) for qubit in self.qubits]
        self.connectivity = ConnectivityDto.from_dict(self.connectivity)
        self.supported_input_formats = [JobInputFormat(input_format) for input_format in self.supported_input_formats]
        self.shots_range = ShotsRangeDto.from_dict(self.shots_range)

    @classmethod
    def from_dict(cls, data: Dict):
        return init_with_defined_params(cls, data)


class AvailabilityTimesDto(BaseModel):
    granularity: str
    start: time
    end: time

    @classmethod
    def from_dict(cls, data: Dict):
        return init_with_defined_params(cls, data)


class CostDto(BaseModel):
    granularity: str
    currency: str
    value: float

    @classmethod
    def from_dict(cls, data: Dict):
        return init_with_defined_params(cls, data)


class BackendStateInfosDto(BaseModel):
    status: BackendStatus
    queue_avg_time: Optional[int] = None
    queue_size: Optional[int] = None
    provider_token_valid: Optional[bool] = None

    def __post_init__(self):
        self.status = BackendStatus(self.status) if self.status else None

    @classmethod
    def from_dict(cls, data: Dict):
        return init_with_defined_params(cls, data)


class BackendDto(BaseModel):
    id: str
    provider: Provider

    @field_validator('provider', mode='before')
    def _validate_provider(cls, v):
        return Provider.from_str(v)

    internal_id: Optional[str] = None
    hardware_provider: Optional[HardwareProvider] = None

    @field_validator('hardware_provider', mode='before')
    def _validate_hardware_provider(cls, v):
        return HardwareProvider.from_str(v)

    name: Optional[str] = None
    documentation: Optional[DocumentationDto] = None
    configuration: Optional[ConfigurationDto] = None
    type: Optional[BackendType] = None
    status: Optional[BackendStatus] = None
    availability: Optional[List[AvailabilityTimesDto]] = None
    costs: Optional[List[CostDto]] = None
    updated_at: Optional[date] = None
    avg_queue_time: Optional[int] = None



from qiskit.providers import Options


class OptionsV2(Options):

    def __getitem__(self, key):
        return super().__getitem__(key)

    def __iter__(self):
        return super().__iter__()

    def __len__(self):
        return super().__len__()

    def __setitem__(self, key, value):
        super().__setitem__(key, value)

    @property
    def __dict__(self):
        return super().__dict__

    def __getattr__(self, name):
        return super().__getattr__(name)

    def __setattr__(self, key, value):
        super().__setattr__(key, value)

    def __getstate__(self):
        return super().__getstate__()

    def __setstate__(self, state):
        super().__setstate__(state)

    def __copy__(self):
        return super().__copy__()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __repr__(self):
        return super().__repr__()

    def __eq__(self, other):
        return super().__eq__(other)

    def set_validator(self, field, validator_value):
        super().set_validator(field, validator_value)

    def update_options(self, **fields):
        super().update_options(**fields)

    def __str__(self):
        return super().__str__()

    @property
    def data(self):
        return list(self.__dict__.keys())

    def update_config(self, **fields):
        super().update_options(**fields)



import json
from typing import Dict
from typing import Optional, List

from qiskit import QuantumCircuit
from qiskit.providers.models import BackendStatus
from qiskit_ibm_provider import IBMBackend
from qiskit_ibm_provider.job import IBMCircuitJob
from qiskit_ibm_provider.utils import RuntimeEncoder

from planqk.client.job_dtos import JobDto, RuntimeJobParamsDto
from planqk.client.model_enums import BackendStatus, JobInputFormat
from planqk.qiskit import PlanqkQiskitJob
from planqk.qiskit.planqk_runtime_job import PlanqkRuntimeJob
from planqk.qiskit.providers.ibm.ibm_backend import PlanqkIbmQiskitBackend


def _encode_circuit_base64(circuit: QuantumCircuit):
    # Transforms circuit to base64 encoded byte stream
    input_json_str = json.dumps(circuit, cls=RuntimeEncoder)
    # Transform back to json but with the base64 encoded byte stream
    return json.loads(input_json_str)


class PlanqkIbmProviderBackend(PlanqkIbmQiskitBackend):

    def __init__(self, **kwargs):
        PlanqkIbmQiskitBackend.__init__(self, **kwargs)
        self.ibm_backend = IBMBackend(configuration=self.configuration(),
                                      provider=kwargs.get('provider'),
                                      api_client=None)
        self.ibm_backend._runtime_run = self._submit_job
        self.ibm_backend.properties = self._backend_properties
        self.ibm_backend.status = self.status

    def run(self, circuit, **kwargs) -> PlanqkRuntimeJob:
        return IBMBackend.run(self.ibm_backend, circuit, **kwargs)

    def status(self):
        operational = self.backend_info.status == BackendStatus.ONLINE
        status_msg = "active" if operational else self.backend_info.status.name.lower()
        pending_jobs = 0  # TODO set pending jobs

        return BackendStatus.from_dict({'backend_name': self.name,
                                        'backend_version': self.backend_version,
                                        'operational': operational,
                                        'status_msg': status_msg,
                                        'pending_jobs': pending_jobs})

    def _submit_job(
        self,
        program_id: str,
        inputs: Dict,
        backend_name: str,
        job_tags: Optional[List[str]] = None,
        image: Optional[str] = None,
    ) -> IBMCircuitJob:
        encoded_input = _encode_circuit_base64(circuit=inputs)
        hgp_name = 'ibm-q/open/main'

        runtime_job_params = RuntimeJobParamsDto(
            program_id=program_id,
            hgp=hgp_name
        )

        job_request = JobDto(backend_id=self.backend_info.id,
                             provider=self.backend_info.provider.name,
                             input_format=JobInputFormat.QISKIT,
                             input=encoded_input,
                             shots=inputs.get('shots'),
                             input_params=runtime_job_params.dict())

        return PlanqkRuntimeJob(backend=self, job_details=job_request, )

    def _backend_properties(self):
        # Supported for loading dynamic properties currently not supported
        return None

    def retrieve_job(self, job_id: str) -> PlanqkQiskitJob:
        job_details = self._planqk_client.get_job(job_id)
        return PlanqkRuntimeJob(backend=self, job_id=job_id, job_details=job_details)



from cmath import exp
from typing import List
from typing import TYPE_CHECKING

from qiskit.circuit import Gate
from qiskit.circuit import Parameter
from qiskit.circuit import QuantumCircuit
from qiskit.circuit.equivalence_library import SessionEquivalenceLibrary
from qiskit.circuit.library import CPhaseGate
from qiskit.quantum_info import Kraus

if TYPE_CHECKING:
    import qiskit


class PCPGate(Gate):
    r"""Implements the phase-shifted controlled-phase gate (PCP).

    This class implements an idealized version of a controlled-phase gate as it can
    potentially be realized by the Rydberg platform. Similarly to the :class:`PCZGate`,
    the gate is hereby only realized up to single-qubit phase gates. The phase shift
    :math:`\theta` depends on the phase :math:`\lambda` of the gate.

    Unitary matrix representation:

    .. math::

        PCP =
        \begin{pmatrix}
        1 & 0 & 0 & 0 \\
        0 & e^{i\theta(\lambda)} & 0 & 0 \\
        0 & 0 & e^{i\theta(\lambda)} & 0 \\
        0 & 0 & 0 & e^{i(2\theta(\lambda)+\lambda)}
        \end{pmatrix}

    Note that the :math:`\lambda`-dependence that is implemented in this class is only
    accurate for :math:`\lambda \leq \pi`.

    .. testcode::

        from qiskit_qryd_provider import PCPGate
        import numpy as np

        assert np.round(PCPGate.get_theta(np.pi), 2) == 2.13

    """

    def __init__(self, lam: float, label: str = None) -> None:
        """Create a new PCP gate.

        Args:
            lam: Phase of the gate.
            label: Optional label for the gate.

        """
        super().__init__("pcp", 2, [lam], label=label)

    def _define(self) -> None:
        """Define the gate."""
        qc = QuantumCircuit(2)
        lam = self.params[0]
        theta = self.get_theta(lam)
        qc.u(0, 0, theta, 0)
        qc.u(0, 0, theta, 1)
        qc.cp(lam, 0, 1)
        self.definition = qc

    def to_matrix(self) -> List[List[complex]]:
        """Return the unitary matrix of the gate.

        Returns:
            A two-dimensional array for the gate unitary matrix.

        """
        lam = self.params[0]
        theta = self.get_theta(lam)
        return [
            [1, 0, 0, 0],
            [0, exp(1j * theta), 0, 0],
            [0, 0, exp(1j * theta), 0],
            [0, 0, 0, exp(2j * self._theta + 1j * lam)],
        ]

    def to_kraus(self) -> "qiskit.circuit.Instruction":
        """Return the Kraus representation of the gate.

        Raises:
            NotImplementedError: If the Kraus representation is not set.

        Returns:
            An instruction encapsulating the Kraus representation.

        """
        if self._kraus is None:
            raise NotImplementedError("The Kraus representation is not implemented.")
        return self._kraus

    @classmethod
    def _init_theta(cls) -> None:
        r"""Initialize the :math:`\lambda`-dependent phase shift of the gate.

        Note that after the initialization, a
        decomposition of the PCP gate is stored to Qiskit's
        SessionEquivalenceLibrary.

        """
        cls._lam = Parameter("lambda")
        cls._theta = (
            5.11382
            - 0.32933
            * (
                1.63085 * cls._lam * cls._lam * (2 * cls._lam).exp()
                + cls._lam
                + 0.02899
            ).log()
        )

        # Reset equivalence library
        default = []
        for c in SessionEquivalenceLibrary.get_entry(CPhaseGate(cls._lam)):
            if not c.get_instructions("pcp"):
                default.append(c)
        SessionEquivalenceLibrary.set_entry(CPhaseGate(cls._lam), default)

        SessionEquivalenceLibrary.set_entry(PCPGate(cls._lam), [])

        # Attach new decomposition to the equivalence library
        def_pcp_cz = QuantumCircuit(2)
        def_pcp_cz.append(PCPGate(cls._lam), [0, 1])
        def_pcp_cz.u(0, 0, -cls._theta, 0)
        def_pcp_cz.u(0, 0, -cls._theta, 1)
        SessionEquivalenceLibrary.add_equivalence(CPhaseGate(cls._lam), def_pcp_cz)

        def_cz_pcp = QuantumCircuit(2)
        def_cz_pcp.append(CPhaseGate(cls._lam), [0, 1])
        def_cz_pcp.u(0, 0, cls._theta, 0)
        def_cz_pcp.u(0, 0, cls._theta, 1)
        SessionEquivalenceLibrary.add_equivalence(PCPGate(cls._lam), def_cz_pcp)

    @classmethod
    def get_theta(cls, lam: float) -> float:
        r"""Get the phase shift of the gate for a given phase :math:`\lambda`.

        Args:
            lam: Phase of the gate.

        Returns:
            Angle of the phase shift.

        """
        return float(cls._theta.assign(cls._lam, lam))

    @classmethod
    def set_kraus(cls, kraus: List[List[List[complex]]] = None) -> None:
        """Set the Kraus representation of the gate.

        Args:
            kraus: A three-dimensional array encoding the Kraus representation of the
                gate.

        """
        if kraus is not None:
            cls._kraus = Kraus(kraus).to_instruction()
        else:
            cls._kraus = None


PCPGate._init_theta()
PCPGate.set_kraus(None)



import json
from typing import Optional, Dict, Set, Union

from pydantic import BaseModel, field_serializer

from planqk.client.model_enums import Provider, JobInputFormat, PlanqkSdkProvider


class JobDto(BaseModel):
    provider: Provider
    shots: int = 1
    backend_id: Optional[str] = None
    id: Optional[str] = None
    provider_job_id: Optional[str] = None
    session_id: Optional[str] = None
    input: Optional[Union[str, Dict]] = None
    input_format: Optional[JobInputFormat] = None
    input_params: Optional[Dict] = None
    error_data: Optional[dict] = None
    started_at: Optional[str] = None
    created_at: Optional[str] = None
    ended_at: Optional[str] = None
    name: Optional[str] = None
    tags: Optional[Set[str]] = None
    sdk_provider: Optional[PlanqkSdkProvider] = None

    @field_serializer('provider')
    def serialize_provider(self, provider: Provider) -> str:
        return provider.value

    @field_serializer('sdk_provider')
    def serialize_sdk_provider(self, sdk_provider: PlanqkSdkProvider) -> str:
        return sdk_provider.value if sdk_provider else None

    def __post_init__(self):
        if self.error_data is not None and isinstance(self.error_data, str):
            self.error_data = json.loads(self.error_data)
        if self.input_params is not None and isinstance(self.input_params, str):
            self.input_params = json.loads(self.input_params)


class JobSummary(BaseModel):
    id: str
    provider: Provider
    backend_id: str
    created_at: Optional[str] = None


class RuntimeJobParamsDto(BaseModel):
    program_id: str
    image: Optional[str] = None
    hgp: Optional[str]
    log_level: Optional[str] = None
    session_id: Optional[str] = None
    max_execution_time: Optional[int] = None
    start_session: Optional[bool] = False
    session_time: Optional[int] = None



import json
import logging
import os
import platform
from abc import ABC, abstractmethod
from json import JSONDecodeError

from planqk.exceptions import CredentialUnavailableError

_PERSONAL_ACCESS_TOKEN_NAME = 'PLANQK_PERSONAL_ACCESS_TOKEN'
_SERVICE_EXECUTION_TOKEN_NAME = 'PLANQK_SERVICE_EXECUTION_TOKEN'
_SERVICE_EXECUTION_TOKEN_NAME_DEPRECATED = 'SERVICE_EXECUTION_TOKEN'
_CONFIG_FILE_PATH = 'PLANQK_CONFIG_FILE_PATH'

logger = logging.getLogger(__name__)


class CredentialProvider(ABC):
    @abstractmethod
    def get_access_token(self) -> str:
        pass


class EnvironmentCredential(CredentialProvider):
    def get_access_token(self) -> str:
        access_token = os.environ.get(_SERVICE_EXECUTION_TOKEN_NAME)
        # backwards compatibility, remove in future
        if not access_token:
            access_token = os.environ.get(_SERVICE_EXECUTION_TOKEN_NAME_DEPRECATED)
        if not access_token:
            access_token = os.environ.get(_PERSONAL_ACCESS_TOKEN_NAME)
        if not access_token:
            message = f'Environment variable {_PERSONAL_ACCESS_TOKEN_NAME} or {_SERVICE_EXECUTION_TOKEN_NAME} not set'
            raise CredentialUnavailableError(message)
        return access_token


def get_config_file_path():
    config_file_path = os.environ.get(_CONFIG_FILE_PATH, None)
    if not config_file_path:
        if platform.system() == 'Windows':
            config_dir = os.path.join(os.getenv('LOCALAPPDATA'), 'planqk')
        else:
            config_dir = os.path.join(os.path.expanduser('~'), '.config', 'planqk')
        config_file_path = os.path.join(config_dir, 'config.json')
    return config_file_path


class ConfigFileCredential(CredentialProvider):
    def __init__(self):
        self.config_file = get_config_file_path()

    def get_access_token(self) -> str:
        if not self.config_file:
            raise CredentialUnavailableError('Config file location not set')
        if not os.path.isfile(self.config_file):
            raise CredentialUnavailableError(f'Config file at {self.config_file} does not exist')
        try:
            access_token = ConfigFileCredential.parse_file(self.config_file)
        except JSONDecodeError:
            raise CredentialUnavailableError('Failed to parse config file: Invalid JSON')
        except KeyError as e:
            raise CredentialUnavailableError(f'Failed to parse config file: Missing expected value - {str(e)}')
        except Exception as e:
            raise CredentialUnavailableError(f'Failed to parse config file: {str(e)}')
        return access_token

    @staticmethod
    def parse_file(path) -> str:
        with open(path, 'r') as file:
            data = json.load(file)
            return data['auth']['value']


class StaticCredential(CredentialProvider):
    def __init__(self, access_token=None):
        self.access_token = access_token

    def get_access_token(self) -> str:
        if not self.access_token:
            raise CredentialUnavailableError(f'Access token not set')
        return self.access_token


class DefaultCredentialsProvider(CredentialProvider):
    def __init__(self, access_token=None):
        self.credentials = [
            StaticCredential(access_token),
            EnvironmentCredential(),
            ConfigFileCredential(),
        ]

    def get_access_token(self) -> str:
        for credential in self.credentials:
            try:
                access_token = credential.get_access_token()
                logger.debug('%s acquired an access token from %s',
                             self.__class__.__name__, credential.__class__.__name__)
                return access_token
            except CredentialUnavailableError:
                logger.debug('%s - %s is unavailable', self.__class__.__name__, credential.__class__.__name__)
            except Exception as e:
                logger.error('%s.get_access_token() failed: %s raised unexpected error "%s"', self.__class__.__name__,
                             credential.__class__.__name__, e, exc_info=logger.isEnabledFor(logging.DEBUG))

        message = f'{self.__class__.__name__} failed to retrieve an access token'
        logger.warning(message)
        raise CredentialUnavailableError(message)



from abc import ABC
from typing import Optional, Any, cast

from braket.circuits import Circuit, Instruction
from braket.circuits.circuit_helpers import validate_circuit_and_shots
from braket.circuits.compiler_directives import StartVerbatimBox
from braket.circuits.gates import PulseGate
from braket.circuits.serialization import QubitReferenceType, OpenQASMSerializationProperties, IRType
from braket.device_schema import GateModelParameters
from braket.device_schema.ionq import IonqDeviceParameters
from braket.device_schema.pulse.frame_v1 import Frame
from braket.device_schema.pulse.port_v1 import Port
from braket.device_schema.rigetti import RigettiDeviceParameters
from braket.device_schema.simulators import GateModelSimulatorDeviceParameters
from braket.error_mitigation import ErrorMitigation
from qiskit import QuantumCircuit

from planqk.backend import PlanqkBackend
from planqk.braket.aws_device import PlanqkAwsDevice
from planqk.braket.planqk_quantum_task import PlanqkAwsQuantumTask
from planqk.client.model_enums import JobInputFormat, HardwareProvider, PlanqkSdkProvider


class PlanqkAwsGateBasedDevice(PlanqkAwsDevice, ABC):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._frames = None
        self._ports = None
        self._gate_calibrations = None

    @property
    def frames(self) -> dict[str, Frame]:
        """Returns a dict mapping frame ids to the frame objects for predefined frames
        for this device.
        """
        self._update_pulse_properties()
        return self._frames or {}

    @property
    def ports(self) -> dict[str, Port]:
        """Returns a dict mapping port ids to the port objects for predefined ports
        for this device.
        """
        self._update_pulse_properties()
        return self._ports or {}

    def run(self, task_specification: QuantumCircuit, shots: Optional[int] = None, *args: Any, **kwargs: Any) -> PlanqkAwsQuantumTask:
        shots = shots if shots else PlanqkAwsDevice.DEFAULT_SHOTS_QPU
        task = PlanqkBackend.run(self, job_input=task_specification, shots=shots, sdk_provider=PlanqkSdkProvider.BRAKET, *args, **kwargs)
        return cast(PlanqkAwsQuantumTask, task)

    def _convert_to_job_input(self, circuit: Circuit, options=None) -> dict:
        """
        Transform the Braket circuit to Braket QASM format.

        Adapted from the Braket SDK's aws_quantum_task.py module.

        Original source:
        Amazon Braket SDK for Python (Apache-2.0 License)
        GitHub Repository: https://github.com/amazon-braket/amazon-braket-sdk-python/blob/main/src/braket/aws/aws_quantum_task.py
        """
        validate_circuit_and_shots(circuit, options.get("shots"))
        disable_qubit_rewiring = options.get("disable_qubit_rewiring", False)
        gate_definitions = options.get("gate_definitions", {})

        qubit_reference_type = QubitReferenceType.VIRTUAL

        if (
            disable_qubit_rewiring
            or Instruction(StartVerbatimBox()) in circuit.instructions
            or gate_definitions
            or any(isinstance(instruction.operator, PulseGate) for instruction in circuit.instructions)
        ):
            qubit_reference_type = QubitReferenceType.PHYSICAL

        serialization_properties = OpenQASMSerializationProperties(
            qubit_reference_type=qubit_reference_type
        )

        openqasm_program = circuit.to_ir(
            ir_type=IRType.OPENQASM,
            serialization_properties=serialization_properties,
            gate_definitions=gate_definitions,
        )

        return openqasm_program.source

    def _circuit_device_params_from_dict(self, device_parameters: dict,
                                         paradigm_parameters: GateModelParameters) -> GateModelSimulatorDeviceParameters:
        if "errorMitigation" in device_parameters:
            error_migitation = device_parameters["errorMitigation"]
            device_parameters["errorMitigation"] = (
                error_migitation.serialize()
                if isinstance(error_migitation, ErrorMitigation)
                else error_migitation
            )
        hardware_provider = self._backend.backend_info.hardware_provider
        if hardware_provider == HardwareProvider.IONQ:
            return IonqDeviceParameters(paradigmParameters=paradigm_parameters, **device_parameters)
        if hardware_provider == HardwareProvider.RIGETTI:
            return RigettiDeviceParameters(paradigmParameters=paradigm_parameters)
        return GateModelSimulatorDeviceParameters(paradigmParameters=paradigm_parameters)

    def _convert_to_job_params(self, job_input: QuantumCircuit = None, options=None) -> dict:
        return {"disable_qubit_rewiring": options.get("disable_qubit_rewiring", False)}

    def _get_job_input_format(self) -> JobInputFormat:
        return JobInputFormat.BRAKET_OPEN_QASM_V3



from qiskit import QuantumCircuit
from qiskit_ionq.helpers import qiskit_circ_to_ionq_circ

from planqk.client.model_enums import JobInputFormat
from planqk.qiskit.options import OptionsV2
from planqk.qiskit.provider import PlanqkQuantumProvider
from planqk.qiskit.providers.azure.azure_backend import PlanqkAzureQiskitBackend


@PlanqkQuantumProvider.register_backend("azure.ionq.simulator")
class PlanqkAzureIonqBackend(PlanqkAzureQiskitBackend):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @classmethod
    def _default_options(cls):
        return OptionsV2(
            shots=500
        )

    def _convert_to_job_input(self, job_input, options=None):
        gateset = options.get("gateset", "qis")
        ionq_circ, _, _ = qiskit_circ_to_ionq_circ(job_input, gateset=gateset)
        return {
            "gateset": gateset,
            "qubits": job_input.num_qubits,
            "circuit": ionq_circ,
        }

    def _get_job_input_format(self) -> JobInputFormat:
        return JobInputFormat.IONQ_CIRCUIT_V1

    def _convert_to_job_params(self, job_input: QuantumCircuit = None, options=None) -> dict:
        memory_option = options.get("memory", None)
        return {"memory": memory_option} if memory_option is not None else {}



import sys

if sys.version_info < (3, 11):
    raise RuntimeError(
        f"PLANQK SDK requires Python 3.11 or higher; "
        f"you are running {sys.version_info.major}.{sys.version_info.minor}."
    )

from planqk.braket.braket_provider import PlanqkBraketProvider
from planqk.qiskit.provider import PlanqkQuantumProvider
from ._version import __version__

__all__ = ['PlanqkQuantumProvider', 'PlanqkBraketProvider']



import json
import warnings
from abc import ABC
from typing import Optional, Union, Any

from braket.annealing import Problem
from braket.aws import AwsDevice, AwsSession
from braket.aws.queue_information import QueueDepthInfo
from braket.circuits import GateCalibrations, Circuit
from braket.device_schema import DeviceCapabilities
from braket.device_schema.pulse.frame_v1 import Frame
from braket.pulse import Port
from networkx import DiGraph

from planqk.backend import PlanqkBackend
from planqk.braket.planqk_quantum_task import PlanqkAwsQuantumTask
from planqk.client.job_dtos import JobDto
from planqk.client.model_enums import Provider, BackendType, BackendStatus, PlanqkSdkProvider
from planqk.decorators import not_implemented, not_supported
from planqk.job import PlanqkBaseJob


class PlanqkAwsDevice(PlanqkBackend, AwsDevice, ABC):

    def __init__(self, **kwargs):
        self._backend_config = None
        super().__init__(**kwargs)

    @property
    def arn(self) -> str:
        warnings.warn("PLANQK AWS devices are identified by their id instead of the ARN.", UserWarning)
        return self.backend_info.internal_id

    @property
    @not_supported
    def aws_session(self) -> AwsSession:
        # Aws session is not supported for PLANQK AWS devices
        pass

    def refresh_metadata(self) -> None:
        """Refresh the `PlanqkAwsDevice` object with the most recent Device metadata."""
        self._backend_info = self._planqk_client.get_backend(backend_id=self.backend_info.id)
        self._get_backend_config(refresh=True)

    @property
    def name(self) -> str:
        return self.backend_info.name if self.backend_info.name is not None else self.backend_info.id

    @property
    def provider_name(self) -> str:
        return self.backend_info.hardware_provider.value

    @property
    def type(self) -> str:
        """str: Return the device type"""
        return "SIMULATOR" if self.backend_info.type == BackendType.SIMULATOR else "QPU"

    @property
    def backend_provider(self) -> str:
        return Provider.AWS.name

    @property
    def provider_name(self) -> str:
        """str: Return the provider name"""
        return self.backend_info.hardware_provider.name

    @property
    @not_implemented
    def ports(self) -> dict[str, Port]:
        """Returns a dict mapping port ids to the port objects for predefined ports
        for this device.
        """
        pass

    @property
    def gate_calibrations(self) -> Optional[GateCalibrations]:
        """Calibration data for a QPU. Calibration data is shown for gates on particular gubits.
        If a QPU does not expose these calibrations, None is returned.

        Returns:
            Optional[GateCalibrations]: The calibration object. Returns `None` if the data
            is not present.
        """
        return None

    @property
    def is_available(self) -> bool:
        """Returns true if the device is currently available.

        Returns:
            bool: Return if the device is currently available.
        """
        planqk_status: BackendStatus = self._get_backend_state().status
        return True if planqk_status == BackendStatus.ONLINE else False

    @property
    def status(self) -> str:
        planqk_status: BackendStatus = self._planqk_client.get_backend_state(self._backend_info.id).status
        return "ONLINE" if planqk_status == BackendStatus.PAUSED else planqk_status.name

    @property
    def properties(self) -> DeviceCapabilities:
        """DeviceCapabilities: Return the device properties

        Please see `braket.device_schema` in amazon-braket-schemas-python_

        .. _amazon-braket-schemas-python: https://github.com/aws/amazon-braket-schemas-python
        """
        config = self._planqk_client.get_backend_config(self._backend_info.id)
        return DeviceCapabilities.parse_raw(json.dumps(config))

    def _run_job(self, job_request: JobDto) -> PlanqkBaseJob:
        job_request.sdk_provider = PlanqkSdkProvider.BRAKET
        return PlanqkAwsQuantumTask(_backend=self, task_id=None, _job_details=job_request, _client=self._planqk_client)

    def run_batch(
        self,
        task_specifications: Union[
            Union[Circuit, Problem],
            list[Union[Circuit, Problem]],
        ],
        shots: Optional[int],
        max_parallel: Optional[int],
        inputs: Optional[Union[dict[str, float], list[dict[str, float]]]],
        *args: Any,
        **kwargs: Any, ):
        raise NotImplementedError("This function is not implemented yet. Please contact PLANQK support if you require this functionality.")

    @property
    def topology_graph(self) -> DiGraph:
        """DiGraph: topology of device as a networkx `DiGraph` object.

        Examples:
            >>> import networkx as nx
            >>> device = AwsDevice("arn1")
            >>> nx.draw_kamada_kawai(device.topology_graph, with_labels=True, font_weight="bold")

            >>> topology_subgraph = device.topology_graph.subgraph(range(8))
            >>> nx.draw_kamada_kawai(topology_subgraph, with_labels=True, font_weight="bold")

            >>> print(device.topology_graph.edges)

        Returns:
            DiGraph: topology of QPU as a networkx `DiGraph` object. `None` if the topology
            is not available for the device.
        """
        return super()._construct_topology_graph()

    @property
    @not_implemented
    def frames(self) -> dict[str, Frame]:
        """Returns a dict mapping frame ids to the frame objects for predefined frames
        for this device.
        """
        pass

    @property
    @not_implemented
    def ports(self) -> dict[str, Port]:
        """Returns a dict mapping port ids to the port objects for predefined ports
        for this device.
        """
        pass

    @not_implemented
    def queue_depth(self) -> QueueDepthInfo:
        """Task queue depth refers to the total number of quantum tasks currently waiting
        to run on a particular device.

        Returns:
            QueueDepthInfo: Instance of the QueueDepth class representing queue depth
            information for quantum tasks and hybrid jobs.
            Queue depth refers to the number of quantum tasks and hybrid jobs queued on a particular
            device. The normal tasks refers to the quantum tasks not submitted via Hybrid Jobs.
            Whereas, the priority tasks refers to the total number of quantum tasks waiting to run
            submitted through Amazon Braket Hybrid Jobs. These tasks run before the normal tasks.
            If the queue depth for normal or priority quantum tasks is greater than 4000, we display
            their respective queue depth as '>4000'. Similarly, for hybrid jobs if there are more
            than 1000 jobs queued on a device, display the hybrid jobs queue depth as '>1000'.
            Additionally, for QPUs if hybrid jobs queue depth is 0, we display information about
            priority and count of the running hybrid job.

        """
        pass

    @not_implemented
    def refresh_gate_calibrations(self) -> Optional[GateCalibrations]:
        """Refreshes the gate calibration data upon request.

        If the device does not have calibration data, None is returned.

        Raises:
            URLError: If the URL provided returns a non 2xx response.

        Returns:
            Optional[GateCalibrations]: the calibration data for the device. None
            is returned if the device does not have a gate calibrations URL associated.
        """
        pass



import random
from collections import defaultdict

import numpy as np
from qiskit.providers import JobStatus
from qiskit.result.models import ExperimentResult

from planqk.qiskit import PlanqkQiskitJob
from planqk.qiskit.providers.azure.result.result_formatter import ResultFormatter


class IonqResultFormatter(ResultFormatter):
    """
       Transform the Azure IonQ results to Qiskit result format.

       Adapted from Azure Quantum Qiskit SDK's job.py module.

       Original source:
       Azure Quantum SDK (MIT License)
       GitHub Repository: https://github.com/microsoft/azure-quantum-python/blob/main/azure-quantum/azure/quantum/qiskit/job.py
    """

    def __init__(self, results: any, job: PlanqkQiskitJob):
        super().__init__(results, job)

    def format_result(self) -> ExperimentResult:
        job_result = {"data": self._format_ionq_results(),
                      "success": True,
                      "header": {},
                      "status": JobStatus.DONE.name,
                      "shots": self.job.shots}

        return ExperimentResult.from_dict(job_result)

    def _format_ionq_results(self) -> dict:
        result = self.results
        job = self.job
        job_details = job._job_details
        num_qubits_str = job_details.input_params.get("qubits")
        if not num_qubits_str:
            raise KeyError(f"Job {job.job_id()} does not have the required metadata (qubits) to format IonQ results.")

        num_qubits = int(num_qubits_str)

        if 'histogram' not in result:
            raise KeyError("Histogram missing in IonQ Job results")

        probabilities = defaultdict(int)
        for key, value in result['histogram'].items():
            bitstring = self._to_bitstring(key, num_qubits)
            probabilities[bitstring] += value

        if job.backend().configuration().simulator:
            counts = self._draw_random_sample(probabilities, job.shots)
        else:
            counts = {bitstring: np.round(self.shots * value) for bitstring, value in probabilities.items()}

        job_details.input_params.get("memory")

        memory = []
        memory_param = job_details.input_params.get("memory")
        if memory_param:
            # Azure Ionq simulator does not support memory natively, hence, it is randomly created. Memory is required for supporting pennylane
            memory = self._generate_random_memory(counts, self.job.shots)

        return {"counts": counts, "probabilities": probabilities, "memory": memory}

    @staticmethod
    def _to_bitstring(k: str, num_qubits):
        # flip bitstring to convert to little Endian
        return format(int(k), f"0{num_qubits}b")[::-1]

    def _draw_random_sample(self, probabilities, shots):
        _norm = sum(probabilities.values())
        if _norm != 1:
            if np.isclose(_norm, 1.0, rtol=1e-2):
                probabilities = {k: v / _norm for k, v in probabilities.items()}
            else:
                raise ValueError(f"Probabilities do not add up to 1: {probabilities}")

        import hashlib
        job_id = self.job.job_id()
        sampler_seed = int(hashlib.sha256(job_id.encode('utf-8')).hexdigest(), 16) % (2 ** 32 - 1)

        rand = np.random.RandomState(sampler_seed)
        rand_values = rand.choice(list(probabilities.keys()), shots, p=list(probabilities.values()))
        return dict(zip(*np.unique(rand_values, return_counts=True)))

    @staticmethod
    def _generate_random_memory(counts: dict, shots: int):
        return random.choices(list(counts.keys()), weights=counts.values(), k=shots)



from .backend import PlanqkQiskitBackend
from .job import PlanqkJob
from .job import PlanqkQiskitJob
from .provider import PlanqkQuantumProvider
from .providers.aws.aws_backend import PlanqkAwsBackend
from .providers.aws.aws_iqm_garnet_backend import PlanqkAwsIqmGarnetBackend
from .providers.aws.aws_rigetti_ankaa_backend import PlanqkAwsRigettiAnkaaBackend
from .providers.azure.ionq_backend import PlanqkAzureIonqBackend
from .providers.qryd.qryd_backend import PlanqkQrydQiskitBackend
from .providers.qudora.qudora_sim_xg1_backend import PlanqkQudoraSimXg1Backend




import json

from braket.device_schema.simulators import GateModelSimulatorDeviceCapabilities

from planqk.braket.braket_provider import PlanqkBraketProvider
from planqk.braket.gate_based_device import PlanqkAwsGateBasedDevice


@PlanqkBraketProvider.register_device("aws.sim.dm1")
class PlanqkAwsDm1Device(PlanqkAwsGateBasedDevice):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @property
    def properties(self) -> GateModelSimulatorDeviceCapabilities:
        """GateModelSimulatorDeviceCapabilities: Return the device properties"""
        config = self._get_backend_config()
        return GateModelSimulatorDeviceCapabilities.parse_raw(json.dumps(config))

    @property
    def name(self) -> str:
        return "dm1"

    @property
    def provider_name(self) -> str:
        return "Amazon Braket"



from __future__ import annotations

from abc import abstractmethod, ABC
from copy import copy
from typing import Union, Optional, Dict, Any, TYPE_CHECKING

from braket.ahs import AnalogHamiltonianSimulation
from braket.circuits import Circuit
from qiskit import QuantumCircuit

from planqk.client.backend_dtos import BackendDto, BackendStateInfosDto
from planqk.client.client import _PlanqkClient
from planqk.client.job_dtos import JobDto
from planqk.client.model_enums import Provider, JobInputFormat, PlanqkSdkProvider
from planqk.job import PlanqkBaseJob

if TYPE_CHECKING:
    from planqk.qiskit.job import PlanqkQiskitJob
    from planqk.braket import PlanqkAwsQuantumTask


class PlanqkBackend(ABC):
    def __init__(  # pylint: disable=too-many-arguments
            self,
            backend_info: BackendDto,
            planqk_client: _PlanqkClient,
    ):
        """PlanqkBackend for executing inputs against PLANQK devices.

        Example:
            provider = PlanqkQuantumProvider()
            backend = provider.get_backend("azure.ionq.simulator")
            input = ... # Can be Qiskit circuit
            backend.run(input, shots=10).result().get_counts()
            {"100": 10, "001": 10}

        Args:
            backend_info: PLANQK actual infos
            provider: Qiskit provider for this actual
            name: name of actual
            description: description of actual
            online_date: online date
            backend_version: actual version
            **fields: other arguments
        """

        self._planqk_client = planqk_client
        if self._planqk_client is None:
            raise RuntimeError("planqk_client must not be None")

        self._backend_info = backend_info

    @property
    def backend_info(self):
        return self._backend_info

    @property
    def min_shots(self):
        return self.backend_info.configuration.shots_range.min

    @property
    def max_shots(self):
        return self.backend_info.configuration.shots_range.max

    def calibration(self) -> Optional[Dict[str, Any]]:
        """
        Retrieve the currently effective calibration data of the backend.

        Note: Backend calibrations are unavailable for simulators.

        Returns:
            Optional[Dict[str, Any]]: Backend calibration data or None if no calibration data is available.
            The data format is backend specific.
        """
        return self._planqk_client.get_backend_calibration(self.backend_info.id)

    def run(self, job_input: Union[QuantumCircuit, Circuit, AnalogHamiltonianSimulation], shots: Optional[int] = None,
            sdk_provider: PlanqkSdkProvider = None,
            **kwargs: object) -> Union[PlanqkQiskitJob, PlanqkAwsQuantumTask]:
        """Runs an experiment on the backend as job.

        Args:
            job_input (QuantumCircuit, AnalogHamiltonianSimulation): job input to run, e.g. a Qiskit circuit or a hamiltonian. Currently only a single input can be executed per job.
            shots (int): the number of shots
            sdk_provider (PlanqkSdkProvider): the SDK provider used for the execution
            **kwargs: additional arguments for the execution
        Returns:
            The job object representing the experiment
        """

        if isinstance(job_input, (list, tuple)):
            if len(job_input) > 1:
                raise ValueError("Multi-experiment jobs are not supported")
            job_input = job_input[0]

        # add kwargs, if defined as options, to a copy of the options
        options = {}
        if hasattr(self, 'options'):
            options = copy(self.options)

        if kwargs:
            for field in kwargs:
                if field in options.data:
                    options[field] = kwargs[field]

        shots = shots if shots else self.backend_info.configuration.shots_range.min
        options['shots'] = shots

        job_input_format = self._get_job_input_format()
        converted_job_input = self._convert_to_job_input(job_input, options)
        input_params = self._convert_to_job_params(job_input, options)

        job_request = JobDto(backend_id=self.backend_info.id,
                             provider=self.backend_info.provider.name,
                             input_format=job_input_format,
                             input=converted_job_input,
                             shots=shots,
                             sdk_provider=sdk_provider,
                             input_params=input_params,
                             name=kwargs.get('name', None))

        return self._run_job(job_request)

    @abstractmethod
    def _convert_to_job_input(self, job_input: Union[QuantumCircuit, AnalogHamiltonianSimulation], options=None) -> dict:
        pass

    def _convert_to_job_params(self, job_input: QuantumCircuit = None, options=None) -> dict:
        return {}

    @abstractmethod
    def _get_job_input_format(self) -> JobInputFormat:
        pass

    @abstractmethod
    def _run_job(self, job_request: JobDto) -> PlanqkBaseJob:
        pass

    @property
    def backend_provider(self) -> Provider:
        """Return the provider offering the quantum backend resource."""
        return self.backend_info.provider

    def _get_backend_config(self, refresh: bool = False) -> Dict[str, Any]:
        if not self._backend_config or refresh == True:
            self._backend_config = self._planqk_client.get_backend_config(self.backend_info.id)
        return self._backend_config

    def _get_backend_state(self) -> BackendStateInfosDto:
        return self._planqk_client.get_backend_state(self.backend_info.id)


from abc import abstractmethod, ABC
from typing import Optional, Union

from qiskit.circuit import Instruction as QiskitInstruction, Delay, Parameter, Reset
from qiskit.circuit import Measure
from qiskit.providers import BackendV2
from qiskit.providers.models import QasmBackendConfiguration, GateConfig
from qiskit.transpiler import Target

from planqk.backend import PlanqkBackend
from planqk.client.backend_dtos import ConfigurationDto, BackendDto, ConnectivityDto
from planqk.client.job_dtos import JobDto
from planqk.client.model_enums import BackendType, PlanqkSdkProvider
from .job import PlanqkQiskitJob
from .options import OptionsV2
from ..client.client import _PlanqkClient


class PlanqkQiskitBackend(PlanqkBackend, BackendV2, ABC):

    def __init__(  # pylint: disable=too-many-arguments
        self,
        planqk_client: _PlanqkClient,
        backend_info: BackendDto,
        **fields,
    ):
        """PlanqkBackend for executing Qiskit circuits against PLANQK devices.

        Example:
            provider = PlanqkQuantumProvider()
            actual = provider.get_backend("azure.ionq.simulator")
            transpiled_circuit = transpile(circuit, actual=actual)
            actual.run(transpiled_circuit, shots=10).result().get_counts()
            {"100": 10, "001": 10}

        Args:
            backend_info: PLANQK actual infos
            provider: Qiskit provider for this actual
            name: name of actual
            description: description of actual
            online_date: online date
            backend_version: actual version
            **fields: other arguments
        """

        PlanqkBackend.__init__(self, planqk_client=planqk_client, backend_info=backend_info)
        BackendV2.__init__(self,
                           provider=backend_info.provider.name,
                           name=backend_info.id,
                           description=f"PLANQK Backend: {backend_info.hardware_provider.name} {backend_info.id}.",
                           online_date=backend_info.updated_at,
                           backend_version="2",
                           **fields)

        self._normalize_qubit_indices()
        self._target = self._planqk_backend_to_target()
        self._configuration = self._planqk_backend_dto_to_configuration()
        self._instance = None

    @abstractmethod
    def _to_gate(self, name: str):
        pass

    @abstractmethod
    def _get_single_qubit_gate_properties(self, instr_name: Optional[str]) -> dict:
        pass

    @abstractmethod
    def _get_multi_qubit_gate_properties(self):
        pass

    non_gate_instr_mapping = {
        "delay": Delay(Parameter("t")),
        "measure": Measure(),
        "reset": Reset(),
    }

    def _normalize_qubit_indices(self):
        # Override to handle non-contiguous zero starting qubit indices
        pass

    @property
    def is_simulator(self):
        return self.backend_info.type == BackendType.SIMULATOR

    def _to_non_gate_instruction(self, name: str) -> Optional[QiskitInstruction]:
        instr = self.non_gate_instr_mapping.get(name, None)
        if instr is not None:
            return instr
        return None

    class PlanqkQiskitTarget(Target):
        def __init__(self, configuration: ConfigurationDto, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._configuration = configuration

        @property
        def physical_qubits(self):
            return sorted(int(qubit.id) for qubit in self._configuration.qubits)

    def _planqk_backend_to_target(self) -> Target:
        """Converts properties of a PLANQK actual into Qiskit Target object.

        Returns:
            target for Qiskit actual
        """

        configuration: ConfigurationDto = self.backend_info.configuration

        self._update_connectivity_with_contiguous_qubit_indices(configuration.connectivity)

        qubit_count: int = self._num_qubits()
        target = self.PlanqkQiskitTarget(configuration=configuration, description=f"Target for PLANQK actual {self.name}", num_qubits=qubit_count)

        single_qubit_props = self._get_single_qubit_gate_properties()
        multi_qubit_props = self._get_multi_qubit_gate_properties()
        gates_names = {gate.name.lower() for gate in configuration.gates}

        for gate_name in gates_names:
            gate = self._to_gate(gate_name)

            if gate is None:
                continue

            if gate.num_qubits == 1:
                target.add_instruction(instruction=gate, properties=single_qubit_props)
            elif gate.num_qubits > 1:
                target.add_instruction(instruction=gate, properties=multi_qubit_props)
            elif gate.num_qubits == 0 and single_qubit_props == {None: None}:
                # For gates without qubit number qargs can not be determined
                target.add_instruction(instruction=gate, properties={None: None})

        measure_props = self._get_single_qubit_gate_properties("measure")
        target.add_instruction(Measure(), measure_props)

        non_gate_instructions = set(configuration.instructions).difference(gates_names).difference({'measure'})
        for non_gate_instruction_name in non_gate_instructions:
            instruction = self._to_non_gate_instruction(non_gate_instruction_name)
            if instruction is not None:
                instr_props = self._get_single_qubit_gate_properties(instruction.name)
                target.add_instruction(instruction, instr_props)

        return target

    def _planqk_backend_dto_to_configuration(self) -> QasmBackendConfiguration:
        basis_gates = [self._get_gate_config_from_target(basis_gate.name)
                       for basis_gate in self.backend_info.configuration.gates if basis_gate.native_gate
                       and self._get_gate_config_from_target(basis_gate.name) is not None]
        gates = [self._get_gate_config_from_target(gate.name)
                 for gate in self.backend_info.configuration.gates if not gate.native_gate
                 and self._get_gate_config_from_target(gate.name) is not None]

        return QasmBackendConfiguration(
            backend_name=self.name,
            backend_version=self.backend_version,
            n_qubits=self._num_qubits(),
            basis_gates=basis_gates,
            gates=gates,
            local=False,
            simulator=self.backend_info.type == BackendType.SIMULATOR,
            conditional=False,
            open_pulse=False,
            memory=self.backend_info.configuration.memory_result_supported,
            max_shots=self.backend_info.configuration.shots_range.max,
            coupling_map=self.coupling_map,
            supported_instructions=self._target.instructions,
            max_experiments=self.backend_info.configuration.shots_range.max,  # Only one circuit is supported per job
            description=self.backend_info.documentation.description,
            min_shots=self.backend_info.configuration.shots_range.min,
            online_date=self.backend_info.updated_at  # TODO replace with online date
        )

    @staticmethod
    def _update_connectivity_with_contiguous_qubit_indices(connectivity: ConnectivityDto) -> None:
        """Device qubit indices may be noncontiguous (label between x0 and x7, x being
        the number of the octagon) while the Qiskit transpiler creates and/or
        handles coupling maps with contiguous indices. This function converts the
        noncontiguous connectivity graph from Aspen to a contiguous one.

        Args:
            connectivity_graph (dict): connectivity graph from Aspen. For example
                4 qubit system, the connectivity graph will be:
                    {"0": ["1", "2", "7"], "1": ["0","2","7"], "2": ["0","1","7"],
                    "7": ["0","1","2"]}

        Returns:
            dict: Connectivity graph with contiguous indices. For example for an
            input connectivity graph with noncontiguous indices (qubit 0, 1, 2 and
            then qubit 7) as shown here:
                {"0": ["1", "2", "7"], "1": ["0","2","7"], "2": ["0","1","7"],
                "7": ["0","1","2"]}
            the qubit index 7 will be mapped to qubit index 3 for the qiskit
            transpilation step. Thereby the resultant contiguous qubit indices
            output will be:
                {"0": ["1", "2", "3"], "1": ["0","2","3"], "2": ["0","1","3"],
                "3": ["0","1","2"]}
        """

        if connectivity.fully_connected:
            return None

        connectivity_graph = connectivity.graph

        # Creates list of existing qubit indices which are noncontiguous.
        indices = sorted(
            int(i)
            for i in set.union(*[{k} | set(v) for k, v in connectivity_graph.items()])
        )
        # Creates a list of contiguous indices for number of qubits.
        map_list = list(range(len(indices)))
        # Creates a dictionary to remap the noncontiguous indices to contiguous.
        mapper = dict(zip(indices, map_list))
        # Performs the remapping from the noncontiguous to the contiguous indices.
        contiguous_connectivity_graph = {
            mapper[int(k)]: [mapper[int(v)] for v in val]
            for k, val in connectivity_graph.items()
        }
        connectivity.graph = contiguous_connectivity_graph

    def _num_qubits(self):
        return self.backend_info.configuration.qubit_count

    def _get_gate_config_from_target(self, name) -> GateConfig:
        operations = [operation for operation in self._target.operations
                      if isinstance(operation.name, str)  # Filters out the IBM conditional instructions having no name
                      and operation.name.casefold() == name.casefold()]
        if len(operations) == 1:
            operation = operations[0]
            return GateConfig(
                name=name,
                parameters=operation.params,
                qasm_def='',
            )

    @property
    def target(self):
        return self._target

    @property
    def max_circuits(self):
        return None

    @classmethod
    def _default_options(cls):
        return OptionsV2()

    def _run_job(self, job_request: JobDto) -> Union[PlanqkQiskitJob]:
        from planqk.qiskit.job_factory import PlanqkQiskitJobFactory

        job_request.sdk_provider = PlanqkSdkProvider.QISKIT
        return PlanqkQiskitJobFactory.create_job(backend=self, job_details=job_request, planqk_client=self._planqk_client)

    def retrieve_job(self, job_id: str) -> PlanqkQiskitJob:
        """Return a single job.

        Args:
            job_id: id of the job to retrieve.

        Returns:
            The job with the given id.
        """
        from planqk.qiskit.job_factory import PlanqkQiskitJobFactory
        return PlanqkQiskitJobFactory.create_job(backend=self, job_id=job_id, planqk_client=self._planqk_client)

    def configuration(self) -> QasmBackendConfiguration:
        """Return the actual configuration.

        Returns:
            QasmBackendConfiguration: the configuration for the actual.
        """
        return self._configuration



def not_implemented(method):
    def wrapper(*args, **kwargs):
        raise NotImplementedError("This function is not implemented. Please contact PLANQK support if you require this functionality.")

    return wrapper


def not_supported(method):
    def wrapper(*args, **kwargs):
        raise NotImplementedError("This function is not supported by the SDK.")

    return wrapper



from typing import Optional

from qiskit.result.models import ExperimentResult

from planqk.client.client import _PlanqkClient
from planqk.client.job_dtos import JobDto
from planqk.client.model_enums import HardwareProvider
from planqk.qiskit import PlanqkQiskitJob, PlanqkQiskitBackend
from planqk.qiskit.providers.azure.result.ionq_result_formatter import IonqResultFormatter
from planqk.qiskit.providers.azure.result.microsoft_v2_result_formatter import MicrosoftV2ResultFormatter


class PlanqkAzureQiskitJob(PlanqkQiskitJob):

    def __init__(self, backend: Optional[PlanqkQiskitBackend], job_id: Optional[str] = None, job_details: Optional[JobDto] = None,
                 planqk_client: Optional[_PlanqkClient] = None):

        super().__init__(backend, job_id, job_details, planqk_client)

    def _create_experiment_result(self, provider_result: dict) -> ExperimentResult:
        backend: PlanqkQiskitBackend = self.backend()
        hw_provider = backend.backend_info.hardware_provider

        if hw_provider == HardwareProvider.IONQ:
            formatter = IonqResultFormatter(provider_result, self)
        elif hw_provider == HardwareProvider.QUANTINUUM:
            formatter = MicrosoftV2ResultFormatter(provider_result, self)
        else:
            raise NotImplementedError(f"Hardware provider {hw_provider} not supported")

        return formatter.format_result()



import logging
from datetime import datetime
from typing import Optional, Union, Callable, Type, Sequence, Dict, List, Any

from qiskit.providers import QiskitBackendNotFoundError
from qiskit_ibm_runtime import RuntimeOptions, ibm_backend
from qiskit_ibm_runtime.accounts import ChannelType
from qiskit_ibm_runtime.utils.result_decoder import ResultDecoder

from planqk.client.job_dtos import RuntimeJobParamsDto, JobDto
from planqk.client.model_enums import Provider
from planqk.qiskit.planqk_runtime_job import PlanqkRuntimeJob
from planqk.qiskit.provider import PlanqkQuantumProvider

logger = logging.getLogger(__name__)


class PlanqkQiskitRuntimeService(PlanqkQuantumProvider):

    def __init__(self, access_token: Optional[str] = None,
                 organization_id: Optional[str] = None,
                 channel: Optional[ChannelType] = None,
                 channel_strategy=None):
        super().__init__(access_token, organization_id)

        self._channel = channel
        self._channel_strategy = channel_strategy
        # Mock close_session method of API client as it is called by the session object after a program has run
        self._api_client = type('Mock_API_Client', (), {'close_session': lambda self: None})

    def backend(
        self,
        name: str = None,
        instance: Optional[str] = None,
    ):
        """Return a single backend matching the specified filtering.

        Args:
            name: Name of the backend.
            instance: This is only supported for ``ibm_quantum`` runtime and is in the
                hub/group/project format. If an instance is not given, among the providers
                with access to the backend, a premium provider will be prioritized.
                For users without access to a premium provider, the default open provider will be used.

        Returns:
            The id of : A backend matching the filtering.

        Raises:
            QiskitBackendNotFoundError: if no backend could be found.
        """
        # Backend returned must be from IBM
        backend = self.get_backend(backend_id=name)
        if backend.backend_provider not in {Provider.IBM, Provider.IBM_CLOUD, Provider.TSYSTEMS}:
            raise QiskitBackendNotFoundError(
                f"Backend '{name}' is not from IBM. Qiskit Runtime only supports IBM backends.")

        return backend

    def _get_backend_object(self, backend_info):
        backend_init_params = {'backend_info': backend_info}
        if backend_info.provider in {Provider.IBM, Provider.IBM_CLOUD, Provider.TSYSTEMS}:
            from planqk.qiskit.providers.ibm.ibm_runtime_backend import PlanqkIbmRuntimeBackend
            return PlanqkIbmRuntimeBackend(planqk_client=self._client, **backend_init_params)
        else:
            return QiskitBackendNotFoundError(
                f"Backends of provider '{backend_info.provider}' are not supported.")

    def run(self,
            program_id: str,
            inputs: Dict,
            options: Optional[Union[RuntimeOptions, Dict]] = None,
            callback: Optional[Callable] = None,
            result_decoder: Optional[Union[Type[ResultDecoder], Sequence[Type[ResultDecoder]]]] = None,
            session_id: Optional[str] = None,
            start_session: Optional[bool] = False, ):  # TODO return planqkjob

        qrt_options: RuntimeOptions = options
        if options is None:
            qrt_options = RuntimeOptions()
        elif isinstance(options, Dict):
            qrt_options = RuntimeOptions(**options)

        qrt_options.validate(channel=self.channel)

        hgp_name = 'ibm-q/open/main'

        runtime_job_params = RuntimeJobParamsDto(
            program_id=program_id,
            image=qrt_options.image,
            hgp=hgp_name,
            log_level=qrt_options.log_level,
            session_id=session_id,
            max_execution_time=qrt_options.max_execution_time,
            start_session=start_session,
            session_time=qrt_options.session_time,
        )

        backend_id = options.get('backend')
        backend = self.backend(backend_id)

        run_options = inputs.get('run_options')
        if run_options is None:
            shots = backend.min_shots
        else:
            shots = run_options.get('shots', backend.min_shots)

        job_input_format = backend._get_job_input_format()
        input_data = backend._convert_to_job_input(inputs)
        input_params = runtime_job_params.dict()

        job_request = JobDto(backend_id=backend_id,
                             provider=Provider.IBM.name,
                             input_format=job_input_format,
                             input=input_data,
                             shots=shots,
                             input_params=input_params)

        return PlanqkRuntimeJob(backend=backend, job_details=job_request, result_decoder=result_decoder, planqk_client=self._client)

    @staticmethod
    def delete_account(
        filename: Optional[str] = None,
        name: Optional[str] = None,
        channel: Optional[ChannelType] = None,
    ) -> bool:
        raise NotImplementedError("Deleting an account is not supported.")

    @staticmethod
    def save_account(
        token: Optional[str] = None,
        url: Optional[str] = None,
        instance: Optional[str] = None,
        channel: Optional[ChannelType] = None,
        filename: Optional[str] = None,
        name: Optional[str] = None,
        proxies: Optional[dict] = None,
        verify: Optional[bool] = None,
        overwrite: Optional[bool] = False,
    ) -> None:
        raise NotImplementedError("Saving an account is not supported.")

    @staticmethod
    def saved_accounts(
        default: Optional[bool] = None,
        channel: Optional[ChannelType] = None,
        filename: Optional[str] = None,
        name: Optional[str] = None,
    ) -> dict:
        raise NotImplementedError("Listing saved accounts is not supported.")

    def job(self, job_id: str) -> PlanqkRuntimeJob:
        """Retrieve a runtime job.

                Args:
                    job_id: Job ID.

                Returns:
                    Runtime job retrieved.
                Raises:
                    PlanqkClientError: If the job cannot be retrieved.
                """
        job_details = self._client.get_job(job_id)
        backend = self.get_backend(job_details.backend_id)

        return PlanqkRuntimeJob(backend=backend, job_id=job_id, job_details=job_details, planqk_client=self._client)

    def jobs(
        self,
        limit: Optional[int] = 10,
        skip: int = 0,
        backend_name: Optional[str] = None,
        pending: bool = None,
        program_id: str = None,
        instance: Optional[str] = None,
        job_tags: Optional[List[str]] = None,
        session_id: Optional[str] = None,
        created_after: Optional[datetime] = None,
        created_before: Optional[datetime] = None,
        descending: bool = True,
    ) -> List[PlanqkRuntimeJob]:
        raise NotImplementedError("Retrieving runtime jobs is not supported.")

    def delete_job(self, job_id: str) -> None:
        """Delete a runtime job.

        Note that this operation cannot be reversed.

        Args:
            job_id: ID of the job to delete.

        Raises:
            PlanqkClientError: If the job cannot be deleted.
        """
        job = self.job(job_id)
        job._cancel()

    def least_busy(
        self,
        min_num_qubits: Optional[int] = None,
        instance: Optional[str] = None,
        filters: Optional[Callable[[List["ibm_backend.IBMBackend"]], bool]] = None,
        **kwargs: Any,
    ) -> ibm_backend.IBMBackend:
        raise NotImplementedError("Retrieving the least busy backend is not supported.")

    @property
    def auth(self) -> str:
        raise NotImplementedError("Retrieving the authentication type is not supported.")

    @property
    def channel(self) -> str:
        """Return the channel type used.

        Returns:
            The channel type used.
        """
        return self._channel

    @property
    def runtime(self):  # type:ignore
        """Return self for compatibility with IBMQ provider.

        Returns:
            self
        """
        return self



import json

from braket.device_schema.ionq import IonqDeviceCapabilities

from planqk.braket.braket_provider import PlanqkBraketProvider
from planqk.braket.gate_based_device import PlanqkAwsGateBasedDevice


@PlanqkBraketProvider.register_device("aws.ionq.aria")
class PlanqkAwsIonqDevice(PlanqkAwsGateBasedDevice):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @property
    def properties(self) -> IonqDeviceCapabilities:
        """IonqDeviceCapabilities: Return the device properties"""
        config = self._get_backend_config()
        return IonqDeviceCapabilities.parse_raw(json.dumps(config))

    @property
    def provider_name(self) -> str:
        return "IonQ"


@PlanqkBraketProvider.register_device("aws.ionq.forte")
class PlanqkAwsIonqForteDevice(PlanqkAwsIonqDevice):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @property
    def name(self) -> str:
        return "Forte 1"


@PlanqkBraketProvider.register_device("aws.ionq.aria")
class PlanqkAwsIonqAriaDevice(PlanqkAwsIonqDevice):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @property
    def name(self) -> str:
        return "Aria 1"


@PlanqkBraketProvider.register_device("aws.ionq.aria-2")
class PlanqkAwsIonq2AriaDevice(PlanqkAwsIonqAriaDevice):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @property
    def name(self) -> str:
        return "Aria 2"



from abc import ABC, abstractmethod
from typing import List, Dict, Any

from planqk.qiskit import PlanqkQiskitJob


class ResultFormatter(ABC):

    def __init__(self, results: any, job: PlanqkQiskitJob):
        self.results = results
        self.job = job

    @abstractmethod
    def format_result(self) -> List[Dict[str, Any]]:
        pass



import json
from typing import Optional, Tuple

from qiskit.circuit import Gate
from qiskit.circuit import IfElseOp, WhileLoopOp, ForLoopOp, SwitchCaseOp, Instruction
from qiskit.circuit import Parameter, Reset
from qiskit.circuit.library import IGate, SXGate, XGate, CXGate, RZGate, ECRGate, CZGate
from qiskit.qobj.utils import MeasLevel, MeasReturnType
from qiskit_ibm_runtime import RuntimeEncoder

from planqk.client.model_enums import JobInputFormat
from planqk.qiskit import PlanqkQiskitBackend
from planqk.qiskit.options import OptionsV2

ibm_name_mapping = {
    "id": IGate(),
    "sx": SXGate(),
    "x": XGate(),
    "cx": CXGate(),
    "rz": RZGate(Parameter("λ")),
    "reset": Reset(),
    "ecr": ECRGate(),
    "cz": CZGate(),
}

qiskit_control_flow_mapping = {
    "if_else": IfElseOp,
    "while_loop": WhileLoopOp,
    "for_loop": ForLoopOp,
    "switch_case": SwitchCaseOp,
}


class PlanqkIbmQiskitBackend(PlanqkQiskitBackend):

    def __init__(self, **kwargs):
        PlanqkQiskitBackend.__init__(self, **kwargs)

    @classmethod
    def _default_options(cls):
        return OptionsV2(
            shots=4000,
            memory=False,
            meas_level=MeasLevel.CLASSIFIED,
            meas_return=MeasReturnType.AVERAGE,
            memory_slots=None,
            memory_slot_size=100,
            rep_time=None,
            rep_delay=None,
            init_qubits=True,
            use_measure_esp=None,
            # Simulator only
            noise_model=None,
            seed_simulator=None,
        )

    def _to_gate(self, name: str) -> Optional[Gate]:
        name = name.lower()
        return ibm_name_mapping.get(name, None) or Gate(name, 0, [])

    def _get_single_qubit_gate_properties(self, instr_name: Optional[str] = None) -> dict:
        qubits = self.backend_info.configuration.qubits
        return {None: None} if self.is_simulator else {(int(qubit.id),): None for qubit in qubits}

    def _get_multi_qubit_gate_properties(self) -> dict:
        connectivity = self.backend_info.configuration.connectivity
        return {None: None} if self.is_simulator else {(int(qubit), int(connected_qubit)): None
                                                       for qubit, connections in connectivity.graph.items()
                                                       for connected_qubit in connections}

    def _to_non_gate_instruction(self, name: str) -> Optional[Instruction]:
        if name in qiskit_control_flow_mapping:
            instr = qiskit_control_flow_mapping[name]
            instr.has_single_gate_props = False
            return instr

        return super()._to_non_gate_instruction(name)

    def _convert_to_job_input(self, job_input, options=None) -> Tuple[JobInputFormat, dict]:
        # Transforms circuit to base64 encoded byte stream
        input_json_str = json.dumps(job_input, cls=RuntimeEncoder)
        # Transform back to json but with the circuit property base64 encoded
        return json.loads(input_json_str)

    def _get_job_input_format(self) -> JobInputFormat:
        return JobInputFormat.QISKIT



import json

from braket.device_schema.iqm import IqmDeviceCapabilities

from planqk.braket.braket_provider import PlanqkBraketProvider
from planqk.braket.gate_based_device import PlanqkAwsGateBasedDevice


@PlanqkBraketProvider.register_device("aws.iqm.garnet")
class PlanqkAwsIqmGarnetDevice(PlanqkAwsGateBasedDevice):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @property
    def name(self) -> str:
        return "Garnet"

    @property
    def provider_name(self) -> str:
        return "IQM"

    @property
    def properties(self) -> IqmDeviceCapabilities:
        """IqmDeviceCapabilities: Return the device properties"""
        config = self._get_backend_config()
        return IqmDeviceCapabilities.parse_raw(json.dumps(config))



import os
from typing import Union

from pydantic import BaseModel, Field

from planqk.credentials import get_config_file_path

_ORGANIZATION_ID = "PLANQK_ORGANIZATION_ID"


class Context(BaseModel):
    id: str = Field(..., description="Id of the user or organization")
    display_name: str = Field(..., alias="displayName", description="Name of the user or organization")
    is_organization: bool = Field(..., alias="isOrganization", description="True if the context is an organization")

    def get_organization_id(self) -> Union[str, None]:
        organization_id = os.environ.get(_ORGANIZATION_ID, None)
        if organization_id:
            return organization_id

        if self.is_organization:
            return self.id
        return None


class Config(BaseModel):
    context: Context = Field(None, description="Context of the user or organization")


class ContextResolver:
    def __init__(self):
        self.config_file = get_config_file_path()

    def get_context(self) -> Union[Context, None]:
        if not os.path.isfile(self.config_file):
            return None

        with open(self.config_file, "r") as f:
            config = Config.model_validate_json(f.read())

        return config.context



import json
import logging
import os
import random
from typing import List, Optional, Callable, Any, Dict

import requests
from requests import Response, HTTPError

from planqk.client.backend_dtos import BackendDto, BackendStateInfosDto
from planqk.client.job_dtos import JobDto
from planqk.client.model_enums import PlanqkJobStatus
from planqk.context import ContextResolver
from planqk.credentials import DefaultCredentialsProvider
from planqk.exceptions import InvalidAccessTokenError, PlanqkClientError, PlanqkError

HEADER_CLOUD_TRACE_CTX = "x-cloud-trace-context"

logger = logging.getLogger(__name__)


def base_url():
    return os.environ.get("PLANQK_QUANTUM_BASE_URL", "https://platform.planqk.de/quantum")


def service_execution_id():
    return os.environ.get("SERVICE_EXECUTION_ID", None)


def _dict_values_to_string(obj_values_dict: dict):
    for key in obj_values_dict:
        obj_value = obj_values_dict[key]
        if not isinstance(obj_value, str):
            str_value = json.dumps(obj_value)
            obj_values_dict[key] = str_value


class _PlanqkClient:
    def __init__(self, access_token: Optional[str] = None, organization_id: Optional[str] = None):
        self._credentials = DefaultCredentialsProvider(access_token)
        self._context_resolver = ContextResolver()
        self._organization_id = organization_id

    def get_credentials(self):
        return self._credentials

    def set_organization_id(self, organization_id: str):
        self._organization_id = organization_id

    def perform_request(self, request_func: Callable[..., Response], url: str, params=None, data=None, headers=None):
        headers = {**self._get_default_headers(), **(headers or {})}
        debug = os.environ.get("PLANQK_QUANTUM_DEBUG", "false").lower() == "true"

        trace_id = headers.get(HEADER_CLOUD_TRACE_CTX, 'unknown')
        try:
            response = request_func(url, json=data, params=params, headers=headers, verify=not debug)
            response.raise_for_status()
            return response.json() if response.status_code != 204 else None
        except requests.exceptions.ConnectionError as e:
            logger.error(f"Cannot connect to middleware under {url} (Trace {trace_id}): {e}")
            raise e
        except HTTPError as e:
            logger.error(f"Request {request_func.__name__} {url} failed (Trace {trace_id}): {e}")
            if e.response.status_code == 401:
                raise InvalidAccessTokenError
            else:
                raise PlanqkClientError(e.response)
        except Exception as e:
            logger.error(f"Request {request_func.__name__} {url} failed (Trace {trace_id}): {e}")
            raise PlanqkError("Error while performing request") from e

    def get_backends(self) -> List[BackendDto]:
        headers = {}
        params = {"onlyPlanqkSdk": True}

        response = self.perform_request(requests.get, f"{base_url()}/backends", params=params, headers=headers)

        return [BackendDto(**backend_info) for backend_info in response]

    def get_backend(self, backend_id: str) -> BackendDto:
        headers = {}

        response = self.perform_request(requests.get, f"{base_url()}/backends/{backend_id}", headers=headers)
        return BackendDto(**response)

    def get_backend_state(self, backend_id: str) -> BackendStateInfosDto:
        headers = {}

        response = self.perform_request(requests.get, f"{base_url()}/backends/{backend_id}/status", headers=headers)
        return BackendStateInfosDto(**response)

    def get_backend_config(self, backend_id: str) -> Dict[str, Any]:
        headers = {}

        response = self.perform_request(requests.get, f"{base_url()}/backends/{backend_id}/config", headers=headers)
        return response

    def get_backend_calibration(self, backend_id: str) -> Dict[str, Any]:
        headers = {}
        response = self.perform_request(requests.get, f"{base_url()}/backends/{backend_id}/calibration", headers=headers)
        return response

    def submit_job(self, job: JobDto) -> JobDto:
        headers = {"content-type": "application/json"}

        # Create dict from job object and remove attributes with None values from it
        job_dict = self._remove_none_values(job.model_dump())

        response = self.perform_request(requests.post, f"{base_url()}/jobs", data=job_dict, headers=headers)
        return JobDto(**response)

    def get_job(self, job_id: str) -> JobDto:
        params = {}
        response = self.perform_request(requests.get, f"{base_url()}/jobs/{job_id}", params=params)
        return JobDto(**response)

    def get_jobs(self) -> List[JobDto]:
        jobs = []
        page = 0
        size = 50

        while True:
            # Request the current page of jobs
            response = self.perform_request(
                requests.get,
                f"{base_url()}/jobs",
                params={"page": page, "size": size}
            )

            # Add the jobs from the current page to the list
            jobs.extend(JobDto(**job_info) for job_info in response.get('content', []))

            # Check if the current page is the last one
            if response.get('last', True):
                break

            page += 1

        return jobs

    def get_job_status(self, job_id: str) -> PlanqkJobStatus:
        response = self.perform_request(requests.get, f"{base_url()}/jobs/{job_id}/status")
        return PlanqkJobStatus[response['status'].upper()]

    def get_job_result(self, job_id: str) -> Dict[str, Any]:
        response = self.perform_request(requests.get, f"{base_url()}/jobs/{job_id}/result")
        return response

    def get_job_calibration(self, job_id: str) -> Dict[str, Any]:
        response = self.perform_request(requests.get, f"{base_url()}/jobs/{job_id}/calibration")
        return response

    def cancel_job(self, job_id: str) -> None:
        self.perform_request(requests.delete, f"{base_url()}/jobs/{job_id}")

    def _get_default_headers(self):
        headers = {"x-auth-token": self._credentials.get_access_token()}

        # inject service execution if present
        if service_execution_id() is not None:
            headers["x-planqk-service-execution-id"] = service_execution_id()

        if self._context_resolver is None:
            self._context_resolver = ContextResolver()

        context = self._context_resolver.get_context()

        if self._organization_id is not None:
            headers["x-organizationid"] = self._organization_id
        elif context is not None and context.is_organization:
            headers["x-organizationid"] = context.get_organization_id()

        headers[HEADER_CLOUD_TRACE_CTX] = self._generate_trace_id()
        logger.debug("PLANQK client request trace id: %s", headers[HEADER_CLOUD_TRACE_CTX])

        return headers

    @classmethod
    def _remove_none_values(cls, d):
        if not isinstance(d, dict):
            return d
        return {k: cls._remove_none_values(v) for k, v in d.items() if v is not None}

    @classmethod
    def _generate_trace_id(cls):
        return '{:032x}'.format(random.getrandbits(128))



from .aws_dm1_device import PlanqkAwsDm1Device
from .aws_sv1_device import PlanqkAwsSv1Device
from .ionq_aria_device import PlanqkAwsIonqDevice
from .iqm_garnet_device import PlanqkAwsIqmGarnetDevice
from .planqk_quantum_task import PlanqkAwsQuantumTask
from .quera_aquila_device import PlanqkQueraAquilaDevice
from .rigetti_ankaa_device import PlanqkAwsRigettiAnkaaDevice



import json

from braket.device_schema.rigetti.rigetti_device_capabilities_v2 import RigettiDeviceCapabilities

from planqk.braket.braket_provider import PlanqkBraketProvider
from planqk.braket.gate_based_device import PlanqkAwsGateBasedDevice


@PlanqkBraketProvider.register_device("aws.rigetti.ankaa")
class PlanqkAwsRigettiAnkaaDevice(PlanqkAwsGateBasedDevice):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @property
    def name(self) -> str:
        return "Ankaa-3"

    @property
    def provider_name(self) -> str:
        return "Rigetti"

    @property
    def properties(self) -> RigettiDeviceCapabilities:
        """RigettiDeviceCapabilities: Return the device properties"""
        config = self._get_backend_config()
        return RigettiDeviceCapabilities.parse_raw(json.dumps(config))



from planqk.qiskit.providers.ibm.ibm_backend import PlanqkIbmQiskitBackend


class PlanqkIbmRuntimeBackend(PlanqkIbmQiskitBackend):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)



from typing import Optional, Dict

from braket.circuits import Circuit, Instruction
from braket.circuits.circuit_helpers import validate_circuit_and_shots
from braket.circuits.compiler_directives import StartVerbatimBox
from braket.circuits.gates import PulseGate
from braket.circuits.serialization import QubitReferenceType, OpenQASMSerializationProperties, IRType
from braket.ir.openqasm import Program as OpenQASMProgram
from qiskit import QuantumCircuit
from qiskit.circuit import Gate
from qiskit.providers import Options
from qiskit_braket_provider.providers.adapter import _GATE_NAME_TO_QISKIT_GATE, to_braket

from planqk.client.model_enums import JobInputFormat
from planqk.qiskit import PlanqkQiskitBackend
from planqk.qiskit.options import OptionsV2
from planqk.qiskit.provider import PlanqkQuantumProvider


@PlanqkQuantumProvider.register_backend("aws.ionq.aria")
@PlanqkQuantumProvider.register_backend("aws.ionq.aria-2")
@PlanqkQuantumProvider.register_backend("aws.ionq.forte")
@PlanqkQuantumProvider.register_backend("aws.sim.dm1")
@PlanqkQuantumProvider.register_backend("aws.sim.sv1")
class PlanqkAwsBackend(PlanqkQiskitBackend):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @classmethod
    def _default_options(cls):
        return OptionsV2()

    @property
    def num_qubits(self) -> int:
        """Return the number of qubits the backend has."""
        # For IQM Garnet backends the qubit size derived from the target is invalid (as they don't use zero-based qubit indices)
        return len(self.backend_info.configuration.qubits)

    def _to_gate(self, name: str) -> Optional[Gate]:
        name = name.lower()
        gate = _GATE_NAME_TO_QISKIT_GATE.get(name, None)
        # Braket quantum backends only support 1 and 2 qubit gates
        return gate if (gate and gate.num_qubits < 3) or self.is_simulator else None

    def _get_single_qubit_gate_properties(self, instr_name: Optional[str] = None) -> dict:
        if self.is_simulator:
            return {None: None}

        qubits = self.backend_info.configuration.qubits
        return {(i,): None for i in range(len(qubits))}

    def _get_multi_qubit_gate_properties(self) -> dict:
        qubits = self.backend_info.configuration.qubits
        connectivity = self.backend_info.configuration.connectivity
        if self.is_simulator:
            return {None: None}
        if connectivity.fully_connected:
            return {(int(qubit1.id), int(qubit2.id)): None for qubit1 in qubits for qubit2 in qubits
                    if qubit1.id != qubit2.id}
        else:
            return {(int(qubit), int(connected_qubit)): None
                    for qubit, connections in connectivity.graph.items()
                    for connected_qubit in connections}

    def _convert_to_job_input(self, job_input: QuantumCircuit, options: Options = None):
        shots = options.get("shots", 1)
        inputs = options.get("inputs", {})
        verbatim = options.get("verbatim", False)

        basis_gates = self.operation_names if not verbatim else None
        braket_circuit = to_braket(job_input, basis_gates, verbatim=verbatim)

        validate_circuit_and_shots(braket_circuit, shots)

        return self._transform_braket_to_qasm_3_program(braket_circuit, False, inputs)

    def _get_job_input_format(self) -> JobInputFormat:
        return JobInputFormat.BRAKET_OPEN_QASM_V3

    def _convert_to_job_params(self, job_input=None, options=None) -> dict:
        return {'disable_qubit_rewiring': False}

    def _transform_braket_to_qasm_3_program(self, braket_circuit: Circuit,
                                            disable_qubit_rewiring: bool,
                                            inputs: Dict[str, float]) -> str:
        """Transforms a Braket input to a QASM 3 program."""

        qubit_reference_type = QubitReferenceType.VIRTUAL

        if (
            disable_qubit_rewiring
            or Instruction(StartVerbatimBox()) in braket_circuit.instructions
            or any(isinstance(instruction.operator, PulseGate) for instruction in braket_circuit.instructions)
        ):
            qubit_reference_type = QubitReferenceType.PHYSICAL

        serialization_properties = OpenQASMSerializationProperties(
            qubit_reference_type=qubit_reference_type
        )

        openqasm_program = braket_circuit.to_ir(
            ir_type=IRType.OPENQASM, serialization_properties=serialization_properties
        )
        if inputs:
            inputs_copy = openqasm_program.inputs.copy() if openqasm_program.inputs is not None else {}
            inputs_copy.update(inputs)
            openqasm_program = OpenQASMProgram(
                source=openqasm_program.source,
                inputs=inputs_copy,
            )

        return openqasm_program.source



import logging
import time
from abc import ABC
from typing import Optional, Dict, Any

from planqk.client.client import _PlanqkClient
from planqk.client.job_dtos import JobDto
from planqk.client.model_enums import PlanqkJobStatus, JOB_FINAL_STATES
from planqk.exceptions import PlanqkError

logger = logging.getLogger(__name__)


class PlanqkBaseJob(ABC):

    def __init__(self, backend: Optional, job_id: Optional[str] = None, job_details: Optional[JobDto] = None,
                 planqk_client: Optional[_PlanqkClient] = None):
        self._client = planqk_client
        if self._client is None:
            raise RuntimeError("planqk_client must not be None")

        if job_id is None and job_details is None:
            raise ValueError("Either 'job_id', 'job_details' or both must be provided.")

        self._result = None
        self._backend = backend
        self._job_details = job_details
        self._status = None

        if job_id is not None and job_details is None:
            self._refresh(job_id)
        elif job_id is None and job_details is not None:
            self._submit()
        else:
            self._job_details = job_details

    def _submit(self):
        """
        Submits the job for execution.
        """

        if self._job_details is None:
            raise RuntimeError("Cannot submit job as no job details are set.")

        self._job_details = self._client.submit_job(self._job_details)

    def _result(self) -> any:
        """Polls for the job result until the job has completed successfully and returns the result.

        Returns:
            The job result if the job completed successfully. The result format is backend specific.

        Raises:
            An error if the job did reach another end state as `JOB_STATUS.COMPLETED`
        """
        if self._result is not None:
            return self._result

        status = self._update_state()
        if status not in JOB_FINAL_STATES:
            status = self._wait_for_final_state()

        if status == PlanqkJobStatus.COMPLETED:
            self._result = self._client.get_job_result(self.id)
            return self._result

        if status == PlanqkJobStatus.FAILED:
            if self._job_details.error_data is None:
                result_data = self._client.get_job_result(self.id)
            self._job_details.error_data = result_data
            raise RuntimeError(
                f"Cannot retrieve results because the job execution failed with status '{status.name}'. "
                f"Reason: {str(result_data)}."
            )

        raise RuntimeError(
            f'{"Cannot retrieve results as job execution has not completed."}'
            + f"(status: {self._status}.")

    def _wait_for_final_state(self, timeout: Optional[float] = None, wait: float = 5) -> PlanqkJobStatus:
        """Poll the job status until it progresses to a final state such as ``DONE`` or ``ERROR``.

        Args:
            timeout: Seconds to wait for the job. If ``None``, wait indefinitely.
            wait: Seconds between queries.
            callback: Callback function invoked after each query.
                The following positional arguments are provided to the callback function:

                * job_id: Job ID
                * job_status: Status of the job from the last query
                * job: This BaseJob instance

                Note: different subclass might provide different arguments to
                the callback function.

        Raises:
            JobTimeoutError: If the job does not reach a final state before the
                specified timeout.
        """
        start_time = time.time()
        status = self._update_state()
        while status not in JOB_FINAL_STATES:
            logger.debug('Waiting for job %s to complete for retrieving its result. Current job status %s', self.id, status)
            elapsed_time = time.time() - start_time
            if timeout is not None and elapsed_time >= timeout:
                raise PlanqkError(f"Timeout while waiting for job {self.id}.")
            time.sleep(wait)
            status = self._update_state()
        return status

    def _update_state(self, use_cached_value: bool = False) -> PlanqkJobStatus:
        """
        Return the status of the job.

        Args:
            use_cached_value (bool): If `True`, uses the value most recently retrieved from PLANQK.
        """
        if not use_cached_value and self._status not in JOB_FINAL_STATES:
            self._status = self._client.get_job_status(self.id)

        return self._status

    @property
    def id(self) -> str:
        """
        This job's id.
        """
        return self._job_details.id if self._job_details is not None else None

    def job_id(self) -> str:
        """
        This job's id.
        """
        return self.id

    @property
    def shots(self) -> int:
        return self._job_details.shots

    def cancel(self):
        """
        Attempt to cancel the job.
        """
        self._client.cancel_job(self.id)

    def _refresh(self, job_id: str = None):
        """
        Refreshes the job details from the server.
        """
        if job_id is None and self.id is None:
            raise ValueError("Job Id is not set.")

        job_id = job_id if job_id is not None else self.id
        self._job_details = self._client.get_job(job_id)
        self._update_state()

    def calibration(self) -> Optional[Dict[str, Any]]:
        """
        Retrieve the backend calibration that was effective when the job execution started.

        Note:
            The backend calibration data is unavailable if the job was executed on a simulator
            or if the job has not been executed yet.

        Returns:
            Optional[Dict[str, Any]]: Backend calibration data or None if no calibration data is available.
            The data format is backend specific.
        """
        return self._client.get_job_calibration(self.id)



from typing import List, Dict

from planqk.client.backend_dtos import QubitDto, ConnectivityDto
from planqk.qiskit.provider import PlanqkQuantumProvider
from planqk.qiskit.providers.aws.aws_backend import PlanqkAwsBackend


def _update_qubits_with_zero_based_ids(qubits: List[QubitDto]):
    for qubit in qubits:
        qubit.id = str(int(qubit.id) - 1)


def _update_connectivity_graph_with_zero_based_qubit_ids(connectivity: ConnectivityDto):
    graph = connectivity.graph
    updated_graph: Dict[str, List[str]] = {}
    for src, targets in graph.items():
        updated_graph[str(int(src) - 1)] = [str(int(target) - 1) for target in targets]

    connectivity.graph = updated_graph


@PlanqkQuantumProvider.register_backend("aws.iqm.garnet")
class PlanqkAwsIqmGarnetBackend(PlanqkAwsBackend):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _normalize_qubit_indices(self):
        _update_qubits_with_zero_based_ids(self.backend_info.configuration.qubits)



from typing import Optional

from qiskit.providers import Backend, JobStatus
from qiskit.result.models import ExperimentResult, ExperimentResultData

from planqk.client.client import _PlanqkClient
from planqk.client.job_dtos import JobDto
from planqk.qiskit import PlanqkQiskitJob


class PlanqkQrydQiskitJob(PlanqkQiskitJob):

    def __init__(self, backend: Optional[Backend], job_id: Optional[str] = None, job_details: Optional[JobDto] = None,
                 planqk_client: Optional[_PlanqkClient] = None):

        super().__init__(backend, job_id, job_details, planqk_client)

    def _create_experiment_result(self, provider_result: dict) -> ExperimentResult:

        if "data" not in provider_result:
            raise KeyError("The expected key 'data' was not found in Qryd job result")

        if "counts" not in provider_result["data"]:
            raise KeyError("The expected key 'counts' was not found in Qryd job result ['data']")

        if "num_qubits" not in provider_result:
            raise KeyError("The expected key 'num_qubits' was not found in Qryd job result")

        num_qubits = provider_result["num_qubits"]
        counts = {self._to_bitstring(hex_str, num_qubits): count for hex_str, count in provider_result["data"]["counts"].items()}

        return ExperimentResult(
            shots=self.shots,
            success=True,
            status=JobStatus.DONE.name,
            data=ExperimentResultData(
                counts=counts,
                memory=[]
            ),
        )

    @staticmethod
    def _to_bitstring(hex_str: str, num_qubits):
        int_value = int(hex_str, 16)
        # flip bitstring to convert to little Endian
        return format(int(int_value), f"0{num_qubits}b")[::-1]



import json
from datetime import datetime
from typing import Optional, Type, Any, Callable, Dict, Union, Sequence

from qiskit.providers import Backend, JobStatus
from qiskit_ibm_runtime import RuntimeJobMaxTimeoutError, RuntimeJobFailureError, RuntimeInvalidStateError
from qiskit_ibm_runtime.constants import DEFAULT_DECODERS
from qiskit_ibm_runtime.utils.result_decoder import ResultDecoder

from planqk.client.client import _PlanqkClient
from planqk.client.job_dtos import JobDto
from planqk.qiskit import PlanqkQiskitJob


class PlanqkRuntimeJob(PlanqkQiskitJob):

    def __init__(self, backend: Optional[Backend], job_id: Optional[str] = None, job_details: Optional[JobDto] = None,
                 result_decoder: Optional[Union[Type[ResultDecoder], Sequence[Type[ResultDecoder]]]] = None,
                 planqk_client: Optional[_PlanqkClient] = None):
        super().__init__(backend, job_id, job_details, planqk_client=planqk_client)
        self._session_id = self._job_details.input_params.get('session_id', None)
        self._program_id = self._job_details.input_params.get('program_id', None)

        decoder = result_decoder or DEFAULT_DECODERS.get(self._program_id, None) or ResultDecoder
        if isinstance(decoder, Sequence):
            self._interim_result_decoder, self._final_result_decoder = decoder
        else:
            self._interim_result_decoder = self._final_result_decoder = decoder

    def interim_results(self, decoder: Optional[Type[ResultDecoder]] = None) -> Any:
        raise NotImplementedError("Interim results are not supported for PLANQK runtime jobs.")

    def _result(  # pylint: disable=arguments-differ
        self,
        timeout: Optional[float] = None,
        decoder: Optional[Type[ResultDecoder]] = None,
    ) -> Any:
        """Return the results of the job.

        Args:
            timeout: Number of seconds to wait for job.
            decoder: A :class:`ResultDecoder` subclass used to decode job results.

        Returns:
            Runtime job result.

        Raises:
            RuntimeJobFailureError: If the job failed.
            RuntimeJobMaxTimeoutError: If the job does not complete within given timeout.
            RuntimeInvalidStateError: If the job was cancelled, and attempting to retrieve result.
        """
        _decoder = decoder or self._final_result_decoder
        if self._result is None or (_decoder != self._final_result_decoder):
            self._wait_for_final_state(timeout=timeout)
            status = self._update_state()
            if status == JobStatus.ERROR:
                error_message = self._reason if self._reason else self._error_message
                if self._reason == "RAN TOO LONG":
                    raise RuntimeJobMaxTimeoutError(error_message)
                raise RuntimeJobFailureError(f"Unable to retrieve job result. {error_message}")
            if status is JobStatus.CANCELLED:
                raise RuntimeInvalidStateError(
                    "Unable to retrieve result for job {}. "
                    "Job was cancelled.".format(self.job_id())
                )

            result_raw = self._client.get_job_result(self.id)

            self._result = _decoder.decode(json.dumps(result_raw)) if result_raw else None
        return self._result

    def stream_results(
        self, callback: Callable, decoder: Optional[Type[ResultDecoder]] = None
    ) -> None:
        raise NotImplementedError("Result streaming is not supported for PLANQK runtime jobs.")

    def cancel_result_streaming(self) -> None:
        raise NotImplementedError("Canceling result streaming is not supported for PLANQK runtime jobs.")

    def logs(self) -> str:
        raise NotImplementedError("Logs are not supported for PLANQK runtime jobs.")

    def metrics(self) -> Dict[str, Any]:
        raise NotImplementedError("Metrics are not supported for PLANQK runtime jobs.")

    @property
    def image(self) -> str:
        """Return the runtime image used for the job.

        Returns:
            Runtime image: image_name:tag or "" if the default
            image is used.
        """
        raise NotImplementedError("Image is not supported for PLANQK runtime jobs.")

    @property
    def inputs(self) -> Dict:
        """Job input parameters.

        Returns:
            Input parameters used in this job.
        """
        raise NotImplementedError("Inputs are not supported for PLANQK runtime jobs.")

    @property
    def program_id(self) -> str:
        """Program ID.

        Returns:
            ID of the program this job is for.
        """
        return self._program_id

    @property
    def creation_date(self) -> Optional[datetime]:
        """Job creation date in local time.

        Returns:
            The job creation date as a datetime object, in local time, or
            ``None`` if creation date is not available.
        """
        raise NotImplementedError("Creation date is not supported for PLANQK runtime jobs.")

    @property
    def session_id(self) -> str:
        """Session ID.

        Returns:
            Job ID of the first job in a runtime session.
        """
        return self._session_id
